#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Description:
    The module sbmlGenerator contains functions to generate sbml files from padmet and txt
    usign the libsbml package

::

    usage:
        sbmlGenerator.py --padmet=FILE --output=FILE --sbml_lvl=STR [--model_id=STR] [--obj_fct=STR] [--mnx_chem_prop=FILE] [--mnx_chem_xref=FILE] [-v]
        sbmlGenerator.py --padmet=FILE --output=FILE [--init_source=STR] [-v]
        sbmlGenerator.py --compound=FILE --output=FILE [--padmetRef=FILE] [-v]
        sbmlGenerator.py --reaction=FILE --output=FILE --padmetRef=FILE [-v]
    
    option:
        -h --help    Show help.
        --padmet=FILE    path of the padmet file to convert into sbml
        --output=FILE    path of the sbml file to generate.
        --mnx_chem_prop=FILE    path of the MNX chemical compounds properties.
        --mnx_chem_xref=FILE    path of the mnx dict of chemical compounds id mapping.
        --reaction=FILE    path of file of reactions ids, one by line to convert to sbml.
        --compound=FILE    path of file of compounds ids, one by line to convert to sbml.
        --init_source=STR    Select the reactions of padmet to convert on sbml based on the source of the reactions, check relations rxn has_reconstructionData.
        --sbml_lvl=STR    sbml level of output.
        --obj_fct=STR    id of the reaction objective.
        -v   print info.
"""
from padmet.classes import PadmetSpec
import padmet.utils.sbmlPlugin as sp
from padmet.utils.gbr import compile_input
import os
import docopt
import libsbml
import re
import subprocess

#default variables
global def_max_lower_bound, def_max_upper_bound, BOUNDARY_ID
def_max_upper_bound = 1000
def_max_lower_bound = -1000
BOUNDARY_ID = 'C-BOUNDARY'

def main():
    args = docopt.docopt(__doc__)
    output = args["--output"]
    verbose = args["-v"]
    if args["--padmet"]:
        padmet = PadmetSpec(args["--padmet"])
        if args["--init_source"]:
            rxn_to_del = set()
            init_source = args["--init_source"].lower()
            rec_rlts = [rlt for rlt in padmet.getAllRelation() if rlt.type == "has_reconstructionData"]
            [rxn_to_del.add(rlt.id_in) for rlt in rec_rlts if padmet.dicOfNode[rlt.id_out].misc.get("SOURCE",[""])[0].lower() == init_source]            
            reactions_to_add = set([node.id for node in list(padmet.dicOfNode.values()) if node.type == "reaction"]).difference(rxn_to_del)
            reaction_to_sbml(reactions_to_add, output, padmet, verbose)
            return
        if args["--model_id"]:
            model_id = args["--model_id"]
        else:
            model_id = os.path.splitext(os.path.basename(output))[0]
        if args["--sbml_lvl"]:
            sbml_lvl = int(args["--sbml_lvl"])
        else:
            sbml_lvl = 3
        obj_fct = args["--obj_fct"]
        mnx_chem_xref = args["--mnx_chem_xref"]
        mnx_chem_prop = args["--mnx_chem_prop"]
        padmet_to_sbml(padmet, output, model_id, obj_fct, sbml_lvl, mnx_chem_prop, mnx_chem_xref, verbose)
    elif args["--reaction"]:
        with open(args["--reaction"], 'r') as f:
            reactions = set(f.read().splitlines())
        padmetRef = PadmetSpec(args["--padmetRef"])
        reaction_to_sbml(reactions, output, padmetRef, verbose)
    elif args["--compound"]:
        with open(args["--compound"], 'r') as f:
            species_compart = [line.split("\t") for line in f.read().splitlines()]
        compound_to_sbml(species_compart, output, verbose)

def padmet_to_sbml(padmet, output, model_id = None, obj_fct = None, sbml_lvl = 3, mnx_chem_prop = None, mnx_chem_xref = None, verbose = False):
    """
    Convert padmet file to sbml file.
    Specificity: 
    - ids are encoded for sbml using functions sbmlPlugin.convert_to_coded_id
    @param padmet_file: the pathname to the padmet file to convert
    @param output: the pathname to the sbml file to create
    @param obj_fct: the identifier of the objection function, the reaction to test in FBA
    @param sbml_lvl: the sbml level
    @param sbml_version: the sbml version
    @param verbose: print informations
    @type padmet_file, output, verbose: str
    @type sbml_lvl, sbml_version: int
    @return: check return of writeSBMLToFile
    @rtype: int
    """
    global all_ga
    #dir_path_gbr = os.path.dirname(os.path.realpath(__file__))+"/grammar-boolean-rapsody.py"
    all_ga = []
    #create an empty sbml model
    with_mnx = False
    if mnx_chem_prop and mnx_chem_xref:
        with_mnx = True
        dict_mnx_chem_xref = parse_mnx_chem_xref(mnx_chem_xref)
        dict_mnx_chem_prop = parse_mnx_chem_prop(mnx_chem_prop)
    if sbml_lvl == 2:
        sbmlns = libsbml.SBMLNamespaces(2,1)
        document = libsbml.SBMLDocument(sbmlns)
        model = document.createModel()
        association = None
        # Create a unit definition
        mmol_per_gDW_per_hr = model.createUnitDefinition()
        check(mmol_per_gDW_per_hr, 'create unit definition')
        check(mmol_per_gDW_per_hr.setId('mmol_per_gDW_per_hr'), 'set unit definition id')
        
        unit = mmol_per_gDW_per_hr.createUnit()
        check(unit, 'create mole unit')
        check(unit.setKind(libsbml.UNIT_KIND_MOLE), 'set unit kind')
        check(unit.setScale(-3), 'set unit scale')
        check(unit.setMultiplier(1), 'set unit multiplier')
        check(unit.setOffset(0), 'set unit offset')
        
        unit = mmol_per_gDW_per_hr.createUnit()
        check(unit, 'create gram unit')
        check(unit.setKind(libsbml.UNIT_KIND_GRAM), 'set unit kind')
        check(unit.setExponent(-1), 'set unit exponent')
        check(unit.setMultiplier(1), 'set unit multiplier')
        check(unit.setOffset(0), 'set unit offset')
        
        unit = mmol_per_gDW_per_hr.createUnit()
        check(unit, 'create second unit')
        check(unit.setKind(libsbml.UNIT_KIND_SECOND), 'set unit kind')
        check(unit.setExponent(-1), 'set unit exponent')
        check(unit.setMultiplier(0.00027777), 'set unit multiplier')
        check(unit.setOffset(0), 'set unit offset')

    elif sbml_lvl == 3:
        sbmlns = libsbml.SBMLNamespaces(3,1,"fbc",1)
        document = libsbml.SBMLDocument(sbmlns)
        document.setPackageRequired("fbc", False)
        model = document.createModel()
        mplugin = model.getPlugin("fbc")
        association = ['<annotation>', 
        '<listOfGeneAssociations xmlns="http://www.sbml.org/sbml/level3/version1/fbc/version1">']
        check(model,                              'create model')
        check(model.setTimeUnits("second"),       'set model-wide time units')
        check(model.setExtentUnits("mole"),       'set model units of extent')
        check(model.setSubstanceUnits('mole'),    'set model substance units')
    if not model_id: model_id = os.path.splitext(os.path.basename(output))[0]
    model.setId(model_id)
    math_ast = libsbml.parseL3Formula('FLUX_VALUE')
    check(math_ast, 'create AST for rate expression')

    #generator of tuple: (x,y) x=species id,y=value of compart, if not defined=""
    species = [(rlt.id_out, rlt.misc.get("COMPARTMENT",[None])[0]) for rlt in padmet.getAllRelation() 
    if rlt.type in ["consumes","produces"]]
    if verbose: print("%s species" %len(species))
    #compart_dict: k = id_encoded, v = original id
    compart_dict = {}    
    #species_dict: k = species_id_encoded, v = dict: k' = {species_id, compart, name}, v' = value or None 
    species_dict = {}
    for species_id, compart in species:
        #encode id for sbml
        species_id_encoded = sp.convert_to_coded_id(species_id, "M", compart)
        #encode compart id for sbml
        #try to get the common_name, if non value return None
        name = padmet.dicOfNode[species_id].misc.get("COMMON-NAME",[species_id])[0]
        #update dicts
        species_dict[species_id_encoded] = {"species_id":species_id, "compart":compart, "name":name}
        
    for species_id_encoded, s_dict in species_dict.items():
        compart = s_dict["compart"]
        name = s_dict["name"]
        original_id = s_dict["species_id"]
        s = model.createSpecies()
        check(s, 'create species')
        check(s.setId(species_id_encoded), 'set species id %s' %species_id_encoded)
        check(s.setMetaId(species_id_encoded), 'set species meta id %s' %species_id_encoded)
        check(s.setBoundaryCondition(False), 'set boundaryCondition to False')
        check(s.setHasOnlySubstanceUnits(False), 'set setHasOnlySubstanceUnits to False')
        check(s.setConstant(False), 'set setConstant to False')
        check(s.setInitialAmount(0.0), 'set initAmount')
        #check(s.setMetaId(metaId), 'set species MetaId %s' %metaId)
        if name is not None:
            check(s.setName(name), 'set species Name %s' %name)
        else:
            check(s.setName(name), 'set species Name %s' %species_id)
        if compart is not None:
            compart_encoded = sp.convert_to_coded_id(compart)
            compart_dict[compart_encoded] = compart
            check(s.setCompartment(compart_encoded), 'set species compartment %s' %compart_encoded)
            if compart == BOUNDARY_ID:
                check(s.setBoundaryCondition(True), 'set boundaryCondition to True')
        if with_mnx:
            try:
                mnx_id = dict_mnx_chem_xref[original_id]
                species_prop = dict(dict_mnx_chem_prop[mnx_id])
            except (IndexError, KeyError) as e:
                #print(species_id)
                species_prop = None
            if species_prop:
                [species_prop.pop(k) for k,v in list(species_prop.items()) if (not v or v == "NA")]
                try:
                    charge = int(species_prop["charge"])
                except (ValueError, KeyError) as e:
                    charge = 0
                formula = species_prop.get("formula","")
                if re.findall("\(|\)|\.",formula): formula = None
                inchi = species_prop.get("inchi", None)
                if sbml_lvl == 3:
                    splugin = s.getPlugin("fbc")
                    check(splugin.setCharge(charge), 'set charge')
                    if formula:
                        check(splugin.setChemicalFormula(formula), 'set Formula')
                    if inchi:
                        annot_xml = create_annotation(inchi, species_id_encoded)
                        check(s.setAnnotation(annot_xml), 'set Annotations')
                    for prop, prop_v in list(species_prop.items()):
                        if prop in ["charge", "formula", "source", "description","inchi"] or prop_v in ["NA",""]:
                            species_prop.pop(prop)
                notes = create_note(species_prop)
                check(s.setNotes(notes), 'set Notes')

    for k, v in compart_dict.items():
        compart = model.createCompartment()
        check(compart,'create compartment')
        check(compart.setId(k),'set compartment id %s' %k)
        check(compart.setSize(1),'set size for compartment id %s' %k)
        check(compart.setConstant(True),'set constant for compartment id %s' %k)

        if v == "c":
            check(compart.setName("cytosol"),'set compartment name cytosol')
        elif v == "e":
            check(compart.setName("extracellular"),'set compartment name extracellular')
        elif v == "p":
            check(compart.setName("periplasm"),'set compartment name periplasm')
        elif v != k:
            check(compart.setName(v),'set compartment id %s' %v)

    if obj_fct is not None:
        obj_fct_encoded = sp.convert_to_coded_id(obj_fct)
        if verbose: print("the objectif reaction is: %s" %(obj_fct_encoded))
    reactions = [node for node in padmet.dicOfNode.values() if node.type == "reaction"]
    nb_reactions = str(len(reactions))    
    # Create reactions
    if verbose: print("%s reactions" %nb_reactions)
    for rNode in reactions:
        rId = rNode.id
        rId_encoded = sp.convert_to_coded_id(rId,"R")
        rName = rNode.misc.get("COMMON-NAME",[rId])[0]
        reaction = model.createReaction()
        check(reaction, 'create reaction')
        check(reaction.setId(rId_encoded), 'set reaction id %s' %rId_encoded)
        if rName is not None:
            check(reaction.setName(rName), 'set reaction name %s' %rName)
        check(reaction.setFast(False), 'set fast')

        #generator of tuple (reactant_id,stoichiometry,compart)
        consumed = ((rlt.id_out, rlt.misc["STOICHIOMETRY"][0], rlt.misc.get("COMPARTMENT",[None])[0]) 
        for rlt in padmet.dicOfRelationIn.get(rId, None) if rlt.type == "consumes")
        #generator of tuple (product_id,stoichiometry,compart)        
        produced = ((rlt.id_out, rlt.misc["STOICHIOMETRY"][0], rlt.misc.get("COMPARTMENT",[None])[0]) 
        for rlt in padmet.dicOfRelationIn.get(rId, None) if rlt.type == "produces")
        #set reversibility
        direction = rNode.misc["DIRECTION"][0]
        if direction == "LEFT-TO-RIGHT":
            reversible = False
        else:
            reversible = True
        check(reaction.setReversible(reversible), 'set reaction reversibility flag %s' %reversible)
        if sbml_lvl == 3:
            bound= mplugin.createFluxBound()
            bound.setReaction(rId_encoded)
            bound.setOperation("lessEqual")
            bound.setValue(def_max_upper_bound)

            bound= mplugin.createFluxBound()
            bound.setReaction(rId_encoded)
            bound.setOperation("greaterEqual")
            if reversible:
                bound.setValue(def_max_lower_bound)
            else:
                bound.setValue(0)
            if rId == obj_fct:
                 objective = mplugin.createObjective()
                 objective.setId("obj1")
                 objective.setType("maximize")
                 mplugin.setActiveObjectiveId("obj1")
                 fluxObjective = objective.createFluxObjective()
                 fluxObjective.setReaction(rId_encoded)
                 fluxObjective.setCoefficient(1)
        elif sbml_lvl == 2:
            kinetic_law = reaction.createKineticLaw()
            check(kinetic_law, 'create kinetic law')
            check(kinetic_law.setMath(math_ast), 'set math on kinetic law')
            #add parameter flux_value
            flux_value_k = kinetic_law.createParameter()
            check(flux_value_k, 'create parameter flux_value_k')
            check(flux_value_k.setId('FLUX_VALUE'), 'set parameter flux_value_k id')
            check(flux_value_k.setValue(0), 'set parameter flux_value_k value')
            check(flux_value_k.setUnits('mmol_per_gDW_per_hr'), 'set parameter flux_value_k units')
            #add parameter upper/lower_bound, lower value depend on reversibility
            upper_bound_k = kinetic_law.createParameter()
            check(upper_bound_k, 'create parameter upper_bound_k')
            check(upper_bound_k.setId('UPPER_BOUND'), 'set parameter upper_bound_k')
            check(upper_bound_k.setValue(def_max_upper_bound),'set parameter upper_bounp_k value')
            check(upper_bound_k.setUnits('mmol_per_gDW_per_hr'), 'set parameter uppper_bound_k units')
    
            if reversible:
                lower_bound_k = kinetic_law.createParameter()
                check(lower_bound_k, 'create parameter lower_bound_k')
                check(lower_bound_k.setId('LOWER_BOUND'), 'set parameter lower_bound_k id')
                check(lower_bound_k.setValue(def_max_lower_bound), 'set parameter lower_bound_k value')
                check(lower_bound_k.setUnits('mmol_per_gDW_per_hr'), 'set parameter lower_bound_k units')
            else:
                lower_bound_k = kinetic_law.createParameter()
                check(lower_bound_k, 'create parameter lower_bound_k')
                check(lower_bound_k.setId('LOWER_BOUND'), 'set parameter lower_bound_k id')
                check(lower_bound_k.setValue(0), 'set parameter lower_bound_k value')
                check(lower_bound_k.setUnits('mmol_per_gDW_per_hr'), 'set parameter lower_bound_k units')
            #objective_coeeficient
            if rId == obj_fct:
                obj_fct_k = kinetic_law.createParameter()
                check(obj_fct_k, 'create parameter obj_fct_k')
                check(obj_fct_k.setId('OBJECTIVE_COEFFICIENT'), 'set parameter obj_fct_k id')
                check(obj_fct_k.setValue(1), 'set parameter obj_fct_k value')
            else:
                obj_fct_k = kinetic_law.createParameter()
                check(obj_fct_k, 'create parameter obj_fct_k')
                check(obj_fct_k.setId('OBJECTIVE_COEFFICIENT'), 'set parameter obj_fct_k id')
                check(obj_fct_k.setValue(0), 'set parameter obj_fct_k value')

        for cId, stoich, compart in consumed:
            cId_encoded = sp.convert_to_coded_id(cId,"M",compart)
            try:
                stoich = float(stoich)
            #for case stoich = n
            except ValueError:
                stoich = float(1)
            species_ref = reaction.createReactant()
            check(species_ref, 'create reactant')
            check(species_ref.setSpecies(cId_encoded), 'assign reactant species %s' %cId_encoded)
            check(species_ref.setStoichiometry(stoich), 'set stoichiometry %s' %stoich)
            check(species_ref.setStoichiometry(stoich), 'set stoichiometry %s' %stoich)
            if sbml_lvl == 3: check(species_ref.setConstant(False), 'set constant %s' %False)

        for pId, stoich, compart in produced:
            pId_encoded = sp.convert_to_coded_id(pId,"M",compart)
            try:
                stoich = float(stoich)
            except ValueError:
                stoich = float(1)
            species_ref = reaction.createProduct()
            check(species_ref, 'create product')
            check(species_ref.setSpecies(pId_encoded), 'assign product species %s' %pId_encoded)
            check(species_ref.setStoichiometry(stoich), 'set stoichiometry %s' %stoich)
            if sbml_lvl == 3: check(species_ref.setConstant(False), 'set constant %s' %False)

        linked_genes = set([rlt.id_out for rlt in padmet.dicOfRelationIn.get(rId, [])
        if rlt.type == "is_linked_to"])
        all_suppData = [padmet.dicOfNode[rlt.id_out] for rlt in padmet.dicOfRelationIn[rId] if rlt.type == "has_suppData"]
        #if rxn has suppdata, check in each suppData, if GENE_ASSOCIATION in misc
        #if run gbr.py to convert the gene assoc to a list of tuple representing the assoc
        #ex: #orignia_la: (a or b) and c => #ga_subsets: [(a,b),(c)]
        #add each ga in ga_subsets in all_ga_subsets
        #for each ga in all_ga_subsets: if len == 1: if the only ga len == 1: just add gene, else create OR structure
        #elif len > 1: create AND structure, then for each GA if len GA == 1: just add gene, else create OR structure
        #if no suppdata, if linked_genes: if len linked_genes == 1: just add gene, else create OR structure
        all_ga_subsets = list()
        if all_suppData:
            for suppData in all_suppData:
                try:
                    original_ga = suppData.misc["GENE_ASSOCIATION"][0]
                    ga_for_gbr = re.sub(r" or " , "|", original_ga)
                    ga_for_gbr = re.sub(r" and " , "&", ga_for_gbr)
                    ga_for_gbr = re.sub(r"\s" , "", ga_for_gbr)
                    #ga_for_gbr = "\"" + ga_for_gbr + "\""
                    if re.findall("\||\&",ga_for_gbr) and len(re.findall("\||\&",ga_for_gbr)) < 100:
                        ga_subsets = []
                        [ga_subsets.append(set(i)) for i in compile_input(ga_for_gbr)]
                        for ga in ga_subsets:
                            if ga not in all_ga_subsets:
                                all_ga_subsets.append(ga)
                except KeyError:
                    pass
        if all_ga_subsets:
            for gene_id in linked_genes:
                if not any([gene_id in ga for ga in all_ga_subsets]):
                    all_ga_subsets.append([gene_id])
        else:
            for gene_id in linked_genes:
                all_ga_subsets.append([gene_id])

        if association:
            if all_ga_subsets:
                add_ga(rId_encoded, all_ga_subsets)
            elif linked_genes:
                add_ga(rId_encoded, all_ga_subsets)
        #set notes
        notes_dict = {}
        if linked_genes:
            notes_dict["GENE_ASSOCIATION"] = " or ".join(["("+" and ".join([i for i in g])+")" for g in all_ga_subsets])

        try:
            categories = set([padmet.dicOfNode[rlt.id_out].misc["CATEGORY"][0] for rlt in padmet.dicOfRelationIn.get(rId,[]) if rlt.type == "has_reconstructionData"])
        except KeyError:
            categories = None
        if categories:
            notes_dict["CATEGORIES"] = " and ".join(categories)
        pathways = set([rlt.id_out for rlt in padmet.dicOfRelationIn.get(rId, [])
        if rlt.type == "is_in_pathway"])
        if len(pathways) != 0:
            notes_dict["SUBSYSTEM"] = " , ".join(pathways)
        if list(notes_dict.keys()):
            notes = create_note(notes_dict)            
            check(reaction.setNotes(notes), 'set notes %s' %notes)



    if all_ga:
        for ga in all_ga:
            association.extend(ga)
        association.extend(['</listOfGeneAssociations>', '</annotation>'])
        association = " ".join(association)
        model.setAnnotation(association)
    if verbose: print("Done, creating sbml file: %s" %output)
    libsbml.writeSBMLToFile(document, output)

def parse_mnx_chem_xref(mnx_chem_xref):
    #k = xref id, v = mnx_id
    dict_mnx_chem_xref = {}
    with open(mnx_chem_xref, 'r') as f:
        for k,v in [line.split("\t")[:2] for line in f.read().splitlines() if not line.startswith("#") and not line.startswith("MNX")]:
            if ":" in k: k = k.split(":")[1]
            dict_mnx_chem_xref[k] = v
    return dict_mnx_chem_xref

def parse_mnx_chem_prop(mnx_chem_prop):
    #k=mnx_id, v = dict of data
    dict_mnx_chem_prop = {}
    with open(mnx_chem_prop, 'r') as f:
        for line in (line.split("\t") for line in f.read().splitlines() if not line.startswith("#")):
            dict_mnx_chem_prop[line[0]] = {"description":line[1],"formula":line[2],"charge":line[3],"mass":line[4],"inchi":line[5],"smiles":line[6],"source":line[7],"inchikey":line[8]}
    return dict_mnx_chem_prop

def create_note(dict_data):
    notes = "<body xmlns=\"http://www.w3.org/1999/xhtml\">"
    for k,v in list(dict_data.items()):
        notes += "<p>"+k+": "+v+"</p>"
    notes += "</body>"
    return notes

def create_annotation(inchi, ref_id):
    """
    dict_data, k = url, v = id
    """
    annotations = '<annotation>\n  <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">\n    <rdf:Description rdf:about="'+ref_id+'">\n      <bqbiol:isVersionOf>\n        <rdf:Bag>\n          <rdf:li rdf:resource="http://identifiers.org/inchi/'+inchi+'"/>\n        </rdf:Bag>\n      </bqbiol:isVersionOf>\n    </rdf:Description>\n  </rdf:RDF>\n</annotation>'
    annot_xml = libsbml.XMLNode.convertStringToXMLNode(annotations)
    return annot_xml
    
def add_ga(rId_encoded, all_ga_subsets):
    """
    if list_ga len == 1: only 1 list of gene: if len of this list is 1: just add gene, else create OR structure
    else: create OR structure, then for each list of gene
    for each ga in list_ga: if len == 1: if the only ga len == 1: just add gene, else create OR structure
    elif len > 1: create AND structure, then for each GA if len GA == 1: just add gene, else create OR structure
    if no suppdata, if linked_genes: if len linked_genes == 1: just add gene, else create OR structure
    """    
    global all_ga
    ga_count = len(all_ga) + 1
    ga = [' <geneAssociation id="ga_'+str(ga_count)+'" reaction="'+rId_encoded+'">']
    if len(all_ga_subsets) == 1:
        uniqu_ga_list = list(all_ga_subsets[0])
        if len(uniqu_ga_list) == 1:
            gene_id = sp.convert_to_coded_id(uniqu_ga_list[0])
            ga.append('<gene reference="'+gene_id+'"/>')
        else:
            ga.append('<or>')
            for gene_id in uniqu_ga_list:
                gene_id = sp.convert_to_coded_id(gene_id)
                ga.append('<gene reference="'+gene_id+'"/>')
            ga.append('</or>')
    else:
        ga.append('<or>')
        for ga_list in all_ga_subsets:
            ga_list = list(ga_list)
            if len(ga_list) == 1:
                gene_id = sp.convert_to_coded_id(ga_list[0])
                ga.append('<gene reference="'+gene_id+'"/>')
            else:
                ga.append('<and>')
                for gene_id in ga_list:
                    gene_id = sp.convert_to_coded_id(gene_id)
                    ga.append('<gene reference="'+gene_id+'"/>')
                ga.append('</and>')
        ga.append('</or>')
    ga.append('</geneAssociation>')
    all_ga.append(ga)        
    

#################################

def reaction_to_sbml(reactions, output, padmetRef, verbose = False):
    """
    convert a list of reactions to sbml format based on a given padmet of reference.
    - ids are encoded for sbml using functions sbmlPlugin.convert_to_coded_id
    @param reactions: list of reactions ids
    @param padmetRef: padmet of reference
    @param output: the pathname to the sbml file to create
    @param sbml_lvl: the sbml level
    @param sbml_version: the sbml version
    @param verbose: print informations
    @type reactions: set
    @type output, verbose: str
    @type padmetRef: <Padmet>
    @type sbml_lvl, sbml_version: int
    @return: check return of writeSBMLToFile
    @rtype: int
    """
    #check if all rxn id are in padmetRef.
    all_rxn = set([k for k,v in padmetRef.dicOfNode.items() if v.type == "reaction"])
    rxn_not_in_ref = reactions.difference(all_rxn)
    if len(rxn_not_in_ref) == len(reactions):
        raise KeyError("None of the reactions is in padmetRef")
    else:
        for rxn_id in rxn_not_in_ref:
            if verbose: print("%s not in padmetRef" % rxn_id)
            reactions.remove(rxn_id)
    padmet = PadmetSpec()
    [padmet.copyNode(padmetRef, rxn_id) for rxn_id in reactions]
    padmet_to_sbml(padmet, output, sbml_lvl=2, verbose = verbose)


def compound_to_sbml(species_compart, output, verbose = False):
    """
    convert a list of compounds to sbml format
    if compart_name is not None, then the compounds id will by: M_originalID_compart_name
    if verbose and specified padmetRef and/or padmetSpec: will check if compounds are in one of the padmet files
    Ids are encoded for sbml using functions sbmlPlugin.convert_to_coded_id
    @param compounds_file: the pathname to the file containing the compounds ids and the compart, line = cpd-id\tcompart.
    @param output: the pathname to the sbml file to create
    @param padmetRef_file: the pathname to the file padmet of reference
    @param padmetRef_file: the pathname to the file padmet of a species
    @param compart_name: the default compart to concatenate
    @param sbml_version: the sbml version
    @param verbose: print informations
    @type compounds_file, output, padmetRef_file, padmetSpec_file, verbose: str
    @type sbml_lvl, sbml_version: int
    @return: check return of writeSBMLToFile
    @rtype: int
    """
    document = libsbml.SBMLDocument(2, 1)
    model = document.createModel()

    if verbose: print("%s species" %len(species_compart))
    for data in species_compart:
        species_id = data[0]
        if len(data) == 1:
            compart = "c"
        else:
            compart = data[1]
        sId_encoded = sp.convert_to_coded_id(species_id,"M",compart)
        s = model.createSpecies()
        check(s, 'create species')
        check(s.setId(sId_encoded), 'set species id')
        check(s.setName(species_id), 'set species name')
        check(s.setCompartment(sp.convert_to_coded_id(compart)), 'set species compartment')
    
    libsbml.writeSBMLToFile(document, output)

def check(value, message):
    """If 'value' is None, prints an error message constructed using
    'message' and then exits with status code 1.  If 'value' is an integer,
    it assumes it is a libSBML return status code.  If the code value is
    LIBSBML_OPERATION_SUCCESS, returns without further action; if it is not,
    prints an error message constructed using 'message' along with text from
    libSBML explaining the meaning of the code, and exits with status code 1.
    """
    if value == None:
        raise SystemExit('LibSBML returned a null value trying to ' + message + '.')
    elif type(value) is int:
        if value == libsbml.LIBSBML_OPERATION_SUCCESS:
            return
        else:
            err_msg = 'Error encountered trying to ' + message + '.' \
                 + 'LibSBML returned error code ' + str(value) + ': "' \
                 + libsbml.OperationReturnValue_toString(value).strip() + '"'
            raise TypeError(err_msg)
    else:
        return


if __name__ == "__main__":
    main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Description:
    Contains all necessary functions to generate wikiPages from a padmet file and update 
    a wiki online. Require WikiManager module (with wikiMate,Vendor)

::

    usage:
        wikiGenerator.py --padmet=FILE/DIR --output=DIR --wiki_id=STR [--database=STR] [--padmetRef=FILE] [--log_file=FILE] [-v]
        wikiGenerator.py --aureme_run=DIR --padmetSpec=ID -v
    
    options:
        -h --help     Show help.
        --padmet=FILE    path to padmet file.
        --output=DIR    path to folder to create with all wikipages in subdir.
        --wiki_id=STR    id of the wiki.
        --padmetRef=FILE    path to padmet of reference, ex: metacyc_xx.padmet, if given, able to calcul pathway rate completion.
        --log_file=FILE    log file from an aureme run, use this file to create a wikipage with all the command used during the aureme run.
        --aureme_run=DIR    can use an aureme run as input, will use from config file information for model_id and log_file and padmetRef.
        -v    print info.
"""
from padmet.classes import PadmetRef
from padmet.classes import PadmetSpec
import os
import shutil
from itertools import chain
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib import colors
import math
import docopt
import re

def main(): 
    #files to upload: folder genomic_data, all sbml in output ortho, annot, external, seeds, targets
    args = docopt.docopt(__doc__)
    padmet = args["--padmet"]
    verbose = args["-v"]
    if args["--padmetRef"]:
        padmetRef = PadmetRef(args["--padmetRef"])
    else:
        padmetRef = None
    wiki_id = args["--wiki_id"]
    output = args["--output"]
    log_file = args["--log_file"]
    database = args["--database"]
    
    wikiGenerator(padmet, output, wiki_id, padmetRef, database, log_file, verbose)

def wikiGenerator(padmet, output, wiki_id, padmetRef=None, database=None, log_file=None, verbose=False):
    """
    
    """
    #if padmet is a directory: padmetFiles = list of padmet path, a padmet file must end with .padmet extension
    #padmetFiles = [/path/padmet_a.padmet, /path/padmet_b.padmet]
    #else: padmet could be many path separated by ';', ex: padmet  = "/path_a/padmet_a.padmet;/path_b/padmet_b.padmet"
    #padmetFiles = [/path_a/padmet_a.padmet, /path_b/padmet_b.padmet]
    if os.path.isdir(padmet):
        path = padmet
        all_files = [i for i in next(os.walk(path))[2] if not i.startswith(".~lock")]
        padmetFiles = [os.path.join(path, i) for i in all_files if i.endswith(".padmet")]
        if len(padmetFiles) == 0:
            raise IOError("No padmet found in %s" %path)
        
    else:
        padmetFiles = padmet.split(";")
    
    #if padmetRef given, extract all pathways associated to reactions (dont take in account pathways linked to pathway)
    #create a dict global_pwy_rxn_dict with k = pathway id, v = set of rxn_id
    global_pwy_rxn_dict = dict()
    if padmetRef:
        for rlt in [rlt for rlt in padmetRef.getAllRelation() if rlt.type == "is_in_pathway" and padmetRef.dicOfNode[rlt.id_in].type == "reaction"]:
            pwy_id = rlt.id_out
            rxn_id = rlt.id_in
            try:
                global_pwy_rxn_dict[pwy_id].add(rxn_id)
            except KeyError:
                global_pwy_rxn_dict[pwy_id] = set([rxn_id])

    #if database given, will add external links to specific database.
    #For example if database == metacyc, then will add external link to page title to metacyc website
    #Work only for 'metacyc' or 'bigg'
    if database: database = database.lower()
    if database == "metacyc":
        ext_link = {"reaction": "http://metacyc.org/META/NEW-IMAGE?object=",
                    "pathway": "http://metacyc.org/META/NEW-IMAGE?object=",
                    "metabolite": "http://metacyc.org/META/NEW-IMAGE?object="}
    elif database == "bigg":
        ext_link = {"reaction":"http://bigg.ucsd.edu/universal/reactions/",
                    "metabolite":"http://bigg.ucsd.edu/universal/metabolites/",
                    "pathway":"http://www.genome.jp/dbget-bin/www_bget?"}
    else:
        ext_link = {}

    #create all folder archi in output folder
    createDirectory(output, verbose)
    #total_padmet_data will store all the information from one to n padmet to finally use this data to create the pages
    """
    total_padmet_data["reaction"] = {'rxn-1':{"basic_attrib": dict(),
                                             "formula":dict(),
                                             "padmet_source": set(),
                                             "gene_assoc": dict(),
                                             "padmet_gene": dict(),
                                             "pathway_assoc":dict(),
                                             "reconstruction_data":dict(),
                                             "xref":dict(),
                                             "misc": dict()}}
    total_padmet_data["gene"] = {'gene-1':{"basic_attrib": dict(),
                                           "padmet_source":set(),
                                           "reac_assoc": dict(),
                                           "pathway_assoc":dict()}
    total_padmet_data["pathway"] = {'pwy-1':{"basic_attrib": dict(),
                                             "padmet_source":set(),
                                             "reac_assoc": dict(),
                                             "completion_rate":dict(),
                                             "total_reac_assoc":set()}
    total_padmet_data["metabolite"] = {'cpd-1':{"basic_attrib": dict(),
                                                "consumed_by": dict(),
                                                "produced_by": dict(),
                                                "unknown_dir_by": dict()}
    total_padmet_data["padmet_source"] = {"reaction":dict(),
                                          "gene": dict(),
                                          "pathway": dict(),
                                          "misc":dict()}
    total_padmet_data["reconstruction"] = {"category":dict(),
                                           "tool": dict(),
                                           "comment": dict(),
                                           "source": dict()}
    """
    total_padmet_data = {"reaction":{},
                         "gene":{},
                         "pathway":{},
                         "metabolite":{},
                         "padmet_source":{},
                         "reconstruction":{"category":dict(), "tool": dict(), "comment": dict(), "source": dict()},
                         "misc":{}
                         }
    #For each padmet file path in padmetFiles, extract all data with extract_padmet_data and add them to the dict total_padmet_data
    for padmetFile in padmetFiles:
        if verbose: print("Reading %s" %padmetFile)
        total_padmet_data = extract_padmet_data(padmetFile, total_padmet_data, global_pwy_rxn_dict, padmetRef, verbose)

    #Use reduce_padmet_data function to compress information. For example if all reactions have same values, remplace the list of padmet to tags
    total_padmet_data = reduce_padmet_data(total_padmet_data, verbose)

    #Create each categories pages
    for rxn_id, rxn_dict_data in total_padmet_data["reaction"].items():
        output_file = os.path.join(output, "reactions", rxn_id.replace("/", "__47__"))
        create_biological_page("reaction", rxn_id, rxn_dict_data, total_padmet_data, ext_link, output_file, padmetRef, verbose)
    for gene_id, gene_dict_data in total_padmet_data["gene"].items():
        output_file = os.path.join(output, "genes", gene_id)
        create_biological_page("gene", gene_id, gene_dict_data, total_padmet_data, ext_link, output_file, padmetRef, verbose)
    for pwy_id, pwy_dict_data in total_padmet_data["pathway"].items():
        output_file = os.path.join(output, "pathways", pwy_id.replace("/", "__47__"))
        create_biological_page("pathway", pwy_id, pwy_dict_data, total_padmet_data, ext_link, output_file, padmetRef, verbose)
    for cpd_id, cpd_dict_data in total_padmet_data["metabolite"].items():
        output_file = os.path.join(output, "metabolites", cpd_id)
        create_biological_page("metabolite", cpd_id, cpd_dict_data, total_padmet_data, ext_link, output_file, padmetRef, verbose)
    for org_id, org_dict_data in total_padmet_data["padmet_source"].items():
        output_file = os.path.join(output, "organisms", org_id)
        create_biological_page("organism", org_id, org_dict_data, total_padmet_data, ext_link, output_file, padmetRef, verbose)
    #Create navigation pages
    navigation_folder = os.path.join(output, "navigation")
    create_navigation_page(total_padmet_data, navigation_folder, verbose)

    if log_file:
        create_log_page(log_file, os.path.join(output, "navigation"))

    """
    #create_venn()
    create_main(wiki_id)
    """

def extract_padmet_data(padmetFile, total_padmet_data, global_pwy_rxn_dict=None, padmetRef=None, verbose=False):
    """
    total_padmet_data: k in ['reaction', 'gene', 'organism', 'pathway', ...]
    if k = 'reaction', v = {'misc':{},'gene_assoc':}
    
    For reaction in padmetFile:
        if reaction_id not in total_padmet_data["reaction"].keys():
            add total_padmet_data["reaction"][reaction_id][padmet_source] = dict()
            
    else, add data only if differents from first
    """
    #name of current padmet
    padmet_source = os.path.basename(padmetFile).replace(".padmet","")
    total_padmet_data["padmet_source"][padmet_source] = {"reaction":dict(), "gene": dict(), "pathway": dict(), "misc":dict()}
    current_padmet_dict = total_padmet_data["padmet_source"][padmet_source]
    padmet = PadmetSpec(padmetFile)

    all_rxns = [node for node in padmet.dicOfNode.values() if node.type == "reaction"]
    count = 0
    
    for rxn_node in all_rxns:
        count += 1
        rxn_id = rxn_node.id
        #rxn_id = '1-ACYLGLYCEROL-3-P-ACYLTRANSFER-RXN'
        #rxn_node = padmet.dicOfNode[rxn_id]
        if verbose:
            print("Extracting reaction:%s/%s\t%s" %(count, len(all_rxns), rxn_id), end='\r')
            
        current_padmet_dict["reaction"][rxn_id] = dict()

        #if first occurrenc, dict instantiation, in total_padmet_data["reaction"]
        if rxn_id not in total_padmet_data["reaction"].keys():
            total_padmet_data["reaction"][rxn_id] = {"basic_attrib": dict(), "formula":dict(), "padmet_source": set(), "gene_assoc": dict(), "padmet_gene": dict(), "pathway_assoc":dict(), "reconstruction_data":dict(), "xref":dict(), "misc": dict()}
            #If padmetRef given, search for external links ex ID in bigg, kegg, etc...
            if padmetRef:
                try:
                    xref_node = [padmetRef.dicOfNode[rlt.id_out] for rlt in padmetRef.dicOfRelationIn[rxn_id] if rlt.type == "has_xref"][0]
                    total_padmet_data["reaction"][rxn_id]["xref"] = dict(xref_node.misc)
                except (IndexError, KeyError) as e:
                    pass

        current_rxn_dict = total_padmet_data["reaction"][rxn_id]
        current_rxn_dict["padmet_source"].add(padmet_source)
        current_rxn_dict["reconstruction_data"][padmet_source] = list()

        #Use update_basic_attrib to extract data from node.misc and add them to 'basic_attrib'
        current_rxn_dict = update_basic_attrib(rxn_node, current_rxn_dict, padmet_source)                    
        list_names = [padmet.dicOfNode[rlt.id_out].misc["LABEL"] for rlt in padmet.dicOfRelationIn[rxn_id] if rlt.type == "has_name"]
        if list_names:
            current_names = set()
            [current_names.update(i) for i in list_names]
            if "synonymous" not in current_rxn_dict["basic_attrib"].keys():
                current_rxn_dict["basic_attrib"]["synonymous"] = dict()
            for name in current_names:
                name = name.lower()
                try:
                    current_rxn_dict["basic_attrib"]["synonymous"][name].add(padmet_source)
                except KeyError:
                    current_rxn_dict["basic_attrib"]["synonymous"][name] = set([padmet_source])

        #get all relations where reaction consumes or produces X
        consume_produce_rlts = [rlt for rlt in padmet.dicOfRelationIn[rxn_id] if rlt.type in ["consumes","produces"]]
        for rlt in consume_produce_rlts:
            cpd_id = rlt.id_out
            cpd_node = padmet.dicOfNode[cpd_id]
            #For each compound, update total_padmet_data["metabolite"]
            if cpd_id not in total_padmet_data["metabolite"].keys():
                total_padmet_data["metabolite"][cpd_id] = {"basic_attrib": dict(), "consumed_by": dict(), "produced_by": dict(), "unknown_dir_by": dict()}
            #extract basic attrib of cpd
            current_cpd_dict = total_padmet_data["metabolite"][cpd_id]
            current_cpd_dict = update_basic_attrib(cpd_node, current_cpd_dict, padmet_source)                    

            #store compound depending on the direction of reaction
            if rxn_node.misc['DIRECTION'][0] == "LEFT-TO-RIGHT":
                direction_str = " '''=>''' "
                if rlt.type == "consumes":
                    try:
                        current_cpd_dict['consumed_by'][rxn_id].add(padmet_source)
                    except KeyError:
                        current_cpd_dict['consumed_by'][rxn_id] = set([padmet_source])
                elif rlt.type == "produces":
                    try:
                        current_cpd_dict['produced_by'][rxn_id].add(padmet_source)
                    except KeyError:
                        current_cpd_dict['produced_by'][rxn_id] = set([padmet_source])

            elif rxn_node.misc['DIRECTION'][0] == "REVERSIBLE":
                direction_str = " '''<=>''' "
                for by_type in ["consumed_by", "produced_by"]:
                    try:
                        current_cpd_dict[by_type][rxn_id].add(padmet_source)
                    except KeyError:
                        current_cpd_dict[by_type][rxn_id] = set([padmet_source])
            else:
                direction_str = " '''=>/<=>''' "
                try:
                    current_cpd_dict['unknown_dir_by'][rxn_id].add(padmet_source)
                except KeyError:
                    current_cpd_dict['unknown_dir_by'][rxn_id] = set([padmet_source])
            

        # Recovering the formula
        #reactants and products: list of str: ['stoichio [[cpd_id]][cpd_compart]',...]
        reactants = set([(rlt.misc["STOICHIOMETRY"][0], rlt.id_out, rlt.misc["COMPARTMENT"][0])
        for rlt in padmet.dicOfRelationIn.get(rxn_id, None) if rlt.type == "consumes"])
        products = set([(rlt.misc["STOICHIOMETRY"][0], rlt.id_out, rlt.misc["COMPARTMENT"][0])
        for rlt in padmet.dicOfRelationIn.get(rxn_id, None) if rlt.type == "produces"])

        reactants_id_str = [reac_tuple[0]+" [["+reac_tuple[1]+"]]["+reac_tuple[2]+"]"
                            for reac_tuple in  sorted(reactants, key=lambda reac_tuple: reac_tuple[1])]
        products_id_str = [prod_tuple[0]+" [["+prod_tuple[1]+"]]["+prod_tuple[2]+"]"
                            for prod_tuple in  sorted(products, key=lambda prod_tuple: prod_tuple[1])]
        #join each list with '+' and direction in center
        formula_id = " '''+''' ".join(reactants_id_str)+direction_str+" '''+''' ".join(products_id_str)
        try:
            current_rxn_dict["formula"][formula_id].add(padmet_source)
        except KeyError:
            current_rxn_dict["formula"][formula_id] = set([padmet_source])
            
        #get all pathways assocaited to the reaction    
        pathways_ids = [rlt.id_out for rlt in padmet.dicOfRelationIn[rxn_id] if rlt.type == "is_in_pathway"]
        for pwy_id in pathways_ids:
            pwy_node = padmet.dicOfNode[pwy_id]
            current_padmet_dict["pathway"][pwy_id] = dict()

            #for each pathway update total_padmet_data["pathway"]
            if pwy_id not in current_rxn_dict["pathway_assoc"].keys():
                current_rxn_dict["pathway_assoc"][pwy_id] = dict()
            if pwy_id not in total_padmet_data["pathway"].keys():
                total_padmet_data["pathway"][pwy_id] = {"basic_attrib": dict(), "padmet_source":set(), "reac_assoc": dict(), "completion_rate":dict(), "total_reac_assoc":set()}
            current_pwy_dict = total_padmet_data["pathway"][pwy_id]
            current_pwy_dict = update_basic_attrib(pwy_node, current_pwy_dict, padmet_source)      
            current_pwy_dict["padmet_source"].add(padmet_source)
            if rxn_id not in current_pwy_dict["reac_assoc"].keys():
                current_pwy_dict["reac_assoc"][rxn_id] = set()

            #recovering the nb of reactions associated to the pathway and the completion rate if global_pwy_rxn_dict given (if padmetRef given)
            total_reac_assoc = set()
            nb_total_reac_assoc = "n.a"
            if global_pwy_rxn_dict:
                if pwy_id in global_pwy_rxn_dict.keys():
                    total_reac_assoc = global_pwy_rxn_dict[pwy_id]
                    nb_total_reac_assoc = len(total_reac_assoc)
                try:
                    current_pwy_dict["completion_rate"][padmet_source].add(rxn_id)
                except KeyError:
                    current_pwy_dict["completion_rate"][padmet_source]= set([rxn_id])

            current_pwy_dict["total_reac_assoc"] = total_reac_assoc
            reac_assoc = [rlt.id_in for rlt in padmet.dicOfRelationOut[pwy_id] if rlt.type == "is_in_pathway"]
            nb_reac_assoc = len(reac_assoc)
            current_rxn_dict["pathway_assoc"][pwy_id][padmet_source] = {"nb_total_rxn": nb_total_reac_assoc, "nb_found_rxn": nb_reac_assoc}
            current_pwy_dict["reac_assoc"][rxn_id].add(padmet_source)
                
        #get all genes associated to the reaction
        linked_rlt = [rlt for rlt in padmet.dicOfRelationIn[rxn_id] if rlt.type == "is_linked_to"]
        if linked_rlt:
            for rlt in linked_rlt:
                gene_id = rlt.id_out
                gene_node = padmet.dicOfNode[gene_id]
                current_padmet_dict["gene"][gene_id] = dict()

                if gene_id not in current_rxn_dict["gene_assoc"].keys():
                    current_rxn_dict["gene_assoc"][gene_id] = {padmet_source: dict()}
                if gene_id not in total_padmet_data["gene"].keys():
                    total_padmet_data["gene"][gene_id] = {"basic_attrib": dict(), "padmet_source":set(), "reac_assoc": dict(), "pathway_assoc":dict()}
                current_gene_dict = total_padmet_data["gene"][gene_id]
                current_gene_dict = update_basic_attrib(gene_node, current_gene_dict, padmet_source)      
                current_gene_dict["padmet_source"].add(padmet_source)
                
                
                for pwy_id, pwy_dict in current_rxn_dict["pathway_assoc"].items():
                    if padmet_source in pwy_dict.keys():
                        pwy_current_source = pwy_dict[padmet_source]
                        if pwy_id not in current_gene_dict["pathway_assoc"].keys():
                            current_gene_dict["pathway_assoc"][pwy_id] = {padmet_source: dict()}
                        current_gene_dict["pathway_assoc"][pwy_id][padmet_source] = dict(pwy_current_source)

                if rxn_id not in current_gene_dict["reac_assoc"].keys():
                    current_gene_dict["reac_assoc"][rxn_id] = {padmet_source: dict()}
                #a is_linked_to rlt have in misc a key "SOURCE:ASSIGNMENT"
                #the value can be only the source, ex: OUTPUT_PANTOGRAPH_X
                #or the source and the known assignment: SILI_ANNOTATION:EC-NUMBER
                sources = rlt.misc["SOURCE:ASSIGNMENT"]
                for source_data in sources:
                    #ValueError: can't split and get 2 value == no known assignment
                    try:
                        source, assignment = source_data.split(":")
                    except ValueError:
                        source = source_data
                        assignment = "n.a"
                    rec_data_node = next(node for node in list(padmet.dicOfNode.values()) if (node.type == "reconstructionData" and node.misc.get("SOURCE",["unknown-source"])[0] == source))
                    category = rec_data_node.misc.get("CATEGORY",["unknown-category"])[0].lower()
                    tool = rec_data_node.misc.get("TOOL",["unknown-tool"])[0].lower()
                    comment = rec_data_node.misc.get("COMMENT",["n.a"])[0].lower()
                    source = source.lower()
                    assignment = assignment.lower()
                    if re.findall("^output_.*_from_", source):
                        source = re.sub("^output_.*_from_", "", source)

                    #special patch for aureme workflow, if source start with 'output_*_"
                    dict_src_data = {"source": source, "assignment":assignment, "tool": tool, "comment": comment}
                    try:
                        current_rxn_dict["gene_assoc"][gene_id][padmet_source][category].append(dict_src_data)
                    except KeyError:
                        current_rxn_dict["gene_assoc"][gene_id][padmet_source][category] = list([dict_src_data])
                    try:
                        current_rxn_dict["padmet_gene"][padmet_source].add(gene_id)
                    except KeyError:
                        current_rxn_dict["padmet_gene"][padmet_source] = set([gene_id])

                    try:
                        current_gene_dict["reac_assoc"][rxn_id][padmet_source][category].append(dict_src_data)
                    except KeyError:
                        current_gene_dict["reac_assoc"][rxn_id][padmet_source][category] = list([dict_src_data])
                        
        reconstruction_data = [padmet.dicOfNode[rlt.id_out] for rlt in padmet.dicOfRelationIn[rxn_id] if rlt.type == "has_reconstructionData"]
        #src_data = {category:{source:{comment:comment, tool:tool}}}
        
        for rec_data_node in reconstruction_data:
            #if found, lower to standardize
            category = rec_data_node.misc.get("CATEGORY",["unknown-category"])[0].lower()
            tool = rec_data_node.misc.get("TOOL",["unknown-tool"])[0].lower()
            if category == "manual" and tool == "unknown-tool":
                tool = "curation"
            comment = rec_data_node.misc.get("COMMENT",["n.a"])[0].lower()
            source = rec_data_node.misc.get("SOURCE",["unknown-source"])[0].lower()
            if re.findall("^output.*_from_", source):
                source = re.sub("^output_.*_from_", "", source)
            if category not in total_padmet_data["reconstruction"]["category"].keys():
                total_padmet_data["reconstruction"]["category"][category] = dict()
            if tool not in total_padmet_data["reconstruction"]["tool"].keys():
                total_padmet_data["reconstruction"]["tool"][tool] = dict()
            if comment not in total_padmet_data["reconstruction"]["comment"].keys():
                total_padmet_data["reconstruction"]["comment"][comment] = dict()
            if source not in total_padmet_data["reconstruction"]["source"].keys():
                total_padmet_data["reconstruction"]["source"][source] = dict()
            dict_src_data = {"category": category, "source": source, "tool": tool, "comment": comment}
            current_rxn_dict["reconstruction_data"][padmet_source].append(dict_src_data)

            try:
                total_padmet_data["reconstruction"]["category"][category][rxn_id].add(padmet_source)
            except KeyError:
                total_padmet_data["reconstruction"]["category"][category][rxn_id] = set([padmet_source])
            try:
                total_padmet_data["reconstruction"]["tool"][tool][rxn_id].add(padmet_source)
            except KeyError:
                total_padmet_data["reconstruction"]["tool"][tool][rxn_id] = set([padmet_source])
            try:
                total_padmet_data["reconstruction"]["comment"][comment][rxn_id].add(padmet_source)
            except KeyError:
                total_padmet_data["reconstruction"]["comment"][comment][rxn_id] = set([padmet_source])
            try:
                total_padmet_data["reconstruction"]["source"][source][rxn_id].add(padmet_source)
            except KeyError:
                total_padmet_data["reconstruction"]["source"][source][rxn_id] = set([padmet_source])

    return total_padmet_data

def update_basic_attrib(node, current_node_dict, padmet_source):
    """
    #TODO
    """
    for tag, list_of_v in node.misc.items():
        tag = tag.lower()
        if tag not in current_node_dict["basic_attrib"].keys():
            current_node_dict["basic_attrib"][tag] = dict()
        global_v = current_node_dict["basic_attrib"][tag]
        for v in list_of_v:
            v = v.lower()
            if tag == "molecular-weight": v = v.replace(" ","")
            try:
                global_v[v].add(padmet_source)
            except KeyError:
                global_v[v] = set([padmet_source])
    return current_node_dict

def reduce_padmet_data(total_padmet_data, verbose=False):
    """
    #TODO
    """
    all_padmet_source = set(total_padmet_data["padmet_source"].keys())
    if len(all_padmet_source) == 1:
        global_val = "_1_IN_1_"
    else:
        global_val = "_1_IN_ALL_"
    for rxn_id, rxn_data in total_padmet_data["reaction"].items():
        basic_attrib_dict = rxn_data["basic_attrib"]
        for tag, dict_val_padmet_source in basic_attrib_dict.items():
            for val, set_padmet_source in dict_val_padmet_source.items():
                if set_padmet_source == all_padmet_source:
                    dict_val_padmet_source[val] = global_val
        formula_dict = rxn_data["formula"]
        if len(formula_dict.keys()) == 1:
            formula_dict[list(formula_dict.keys())[0]] = global_val
            
                        

    for cpd_id, cpd_data in total_padmet_data["metabolite"].items():
        basic_attrib_dict = cpd_data["basic_attrib"]
        for tag, dict_val_padmet_source in basic_attrib_dict.items():
            for val, set_padmet_source in dict_val_padmet_source.items():
                if set_padmet_source == all_padmet_source:
                    dict_val_padmet_source[val] = global_val
        for key in ["consumed_by", "produced_by", "unknown_dir_by"]:
            set_padmet_source = cpd_data[key]
            if set_padmet_source == all_padmet_source:
                cpd_data[key] = global_val

    for gene_id, gene_data in total_padmet_data["gene"].items():
        basic_attrib_dict = gene_data["basic_attrib"]
        for tag, dict_val_padmet_source in basic_attrib_dict.items():
            for val, set_padmet_source in dict_val_padmet_source.items():
                if set_padmet_source == all_padmet_source:
                    dict_val_padmet_source[val] = global_val
        
    for pwy_id, pwy_data in total_padmet_data["pathway"].items():
        basic_attrib_dict = pwy_data["basic_attrib"]
        for tag, dict_val_padmet_source in basic_attrib_dict.items():
            for val, set_padmet_source in dict_val_padmet_source.items():
                if set_padmet_source == all_padmet_source:
                    dict_val_padmet_source[val] = global_val
        reac_assoc_dict = pwy_data["reac_assoc"]
        for rxn_id, set_padmet_source in reac_assoc_dict.items():
            if set_padmet_source == all_padmet_source:
                reac_assoc_dict[rxn_id] = global_val
        completion_rate_dict = pwy_data["completion_rate"]
        for padmet_source in all_padmet_source:
            try:
                rxn_assoc = completion_rate_dict[padmet_source]
            except KeyError:
                rxn_assoc = []

            try:
                rate = round(float(len(rxn_assoc))/float(len(pwy_data["total_reac_assoc"])),2)
            except ZeroDivisionError:
                rate = None
            completion_rate_dict[padmet_source] = rate
            
            

    return total_padmet_data
                    
    

def createDirectory(output, verbose=False):
    """
    create the folders genes, reactions, metabolites, pathways in the folder dirPath/
    if already exist, it will replace old folders (and delete old files)

    Parameters
    ----------
    output: str
        path to output folder
    """
    #simple check that dirPath is a dir:
    dirNames = ["genes","reactions","metabolites","pathways","organisms","navigation","files","misc"]
    #creatings the directory which will contains the wiki pages
    for d in dirNames:
        d_path = os.path.join(output,d)
        if not os.path.exists(d_path):
            if verbose: print("Creating directory: %s" %d_path)
            os.makedirs(d_path)
        else:
            if verbose: print("The directory %s already exist. Old pages will be deleted" %d_path)
            shutil.rmtree(d_path)
            os.makedirs(d_path)
    
    

def create_biological_page(category, page_id, page_dict_data, total_padmet_data, ext_link, output_file, padmetRef=None, verbose=False):
    """
    #TODO
    """
    total_padmet_source = list(total_padmet_data["padmet_source"].keys())
    to_collapse_cutoff = 20
    if verbose: print("%s: %s" %(category, page_id), end='\r')
    #stock in properties: all properties associated to the current page for semantic mediawiki
    #properties = [{{#set PROPERTY_X:VALUE_1|...|VALUE_N}}, ...]
    properties = []

    #ext_link is used to create external link to the database of reference if known
    if ext_link.get(category):
        dataInArray = ['[[Category:'+category+']]',
        '== '+category.capitalize()+' ['+ext_link.get(category)+page_id+' '+page_id+'] ==']
    else:
        dataInArray = ['[[Category:'+category+']]',
        '== '+category.capitalize()+' '+page_id+' ==']

    #basic_attrib contains data from misc dict of nodes ex for reaction: common-name, ec-number ...
    #page_dict_data["basic_attrib"] = {tag*: dict_val_padmet_source}
    #dict_va_padmet_source = {value: set_padmet_source}
    #ex: {'ec-number':{'ec-1.2.2: set([padmet_a, padmet_b])}}
    if "basic_attrib" in page_dict_data.keys():
        for tag, dict_val_padmet_source in page_dict_data["basic_attrib"].items():
            line = '* '+tag+':'
            dataInArray.append(line)
            #dont add smile as propertie because of invalid characters for SemanticMediaWiki
            if tag not in ["smiles"]:
                add_property(properties, tag, dict_val_padmet_source.keys())
            for val, set_padmet_source in dict_val_padmet_source.items():
                #for ec-number, remove 'ec-' and add external link to expasy
                if tag == "ec-number":
                    line = "** [http://enzyme.expasy.org/EC/"+val.replace("ec-","")+" "+val+"]"
                #for taxonomic-range from metacyc, add external link to biocyc
                elif tag == "taxonomic-range" and ext_link.get(category):
                    line = "** ["+ext_link.get(category)+val+" "+val+"]"
                #for inchi-key, remove "inchikey="
                elif tag == "inchi-key":
                    line = "** %s" %(val.replace("inchikey=",""))
                else:
                    line = "** %s" %val
                dataInArray.append(line)
                #when set_padmet_source == '_1_IN_ALL_': code means that the value is in all padmet/organisms
                if set_padmet_source == "_1_IN_ALL_":
                    line = "*** Value found in all organisms"
                    dataInArray.append(line)
                #when set_padmet_source != '_1_IN_ALL_' or '_1_IN_1': its a real set of padmet sources
                if set_padmet_source not in ["_1_IN_ALL_", "_1_IN_1_"]:
                    
                    line = "*** Value found in %s organism(s): %s" %(len(set_padmet_source), ";".join(["[[%s]]"%src for src in set_padmet_source]))
                    dataInArray.append(line)

    #For each category, extract in a specific way the information

    if category == "reaction":
        dataInArray.append("== Reaction formula ==")
        #add_property(properties, "formula", page_dict_data["formula"].keys())
        for formula, set_padmet_source in page_dict_data["formula"].items():
            line = "* "+formula
            dataInArray.append(line)
            if set_padmet_source == "_1_IN_ALL_":
                line = "** Same formula in all organisms"
                dataInArray.append(line)
            if set_padmet_source not in ["_1_IN_ALL_", "_1_IN_1_"]:
                line = "** Formula found in %s organism(s): %s" %(len(set_padmet_source), ";".join(set_padmet_source))
                dataInArray.append(line)

        if len(total_padmet_source) > 1:
            add_property(properties, "nb organism associated", [len(page_dict_data["padmet_source"])])
            min_gene_assoc = min([len(genes_set) for genes_set in page_dict_data["padmet_gene"].values()])
            max_gene_assoc = max([len(genes_set) for genes_set in page_dict_data["padmet_gene"].values()])
            avg_gene_assoc = round(sum([len(genes_set) for genes_set in page_dict_data["padmet_gene"].values()])/len(page_dict_data["padmet_gene"].values()),2)
            add_property(properties, "min gene associated", [min_gene_assoc])
            add_property(properties, "max gene associated", [max_gene_assoc])
            add_property(properties, "avg gene associated", [avg_gene_assoc])
            dataInArray.append('== Organism(s) associated with this reaction  ==')
            [dataInArray.append("* [["+padmet_source+"]]") for padmet_source in page_dict_data["padmet_source"]]
        else:
            add_property(properties, "nb gene associated", [len(page_dict_data["gene_assoc"].keys())])
            
        dataInArray.append('== Gene(s) associated with this reaction  ==')
        for gene_id, gene_dict in page_dict_data["gene_assoc"].items():
            line = "* Gene: [[%s]]" %gene_id
            dataInArray.append(line)
            if len(total_padmet_source) > 1:
                for padmet_source, category_dict in gene_dict.items():
                    line = "** In organism: [[%s]]" %padmet_source
                    dataInArray.append(line)
                    for category, category_data in category_dict.items():
                        line = "*** Category: [[%s]]" %category
                        dataInArray.append(line)
                        for data in category_data:
                            line = "**** Source: [[%s]], Tool: [[%s]], Assignment: %s, Comment: %s" %(data["source"], data["tool"], data["assignment"], data["comment"])
                            dataInArray.append(line)
            else:
                for category_dict in gene_dict.values():
                    for category, category_data in category_dict.items():
                        line = "** Category: [[%s]]" %category
                        dataInArray.append(line)
                        for data in category_data:
                            line = "*** Source: [[%s]], Tool: [[%s]], Assignment: %s, Comment: %s" %(data["source"], data["tool"], data["assignment"], data["comment"])
                            dataInArray.append(line)
        
        dataInArray.append('== Pathway(s)  ==')
        #set of pathways id associated to the reaction
        add_property(properties, "nb pathway associated", [len(page_dict_data["pathway_assoc"].keys())])
        for pwy_id, pwy_dict in page_dict_data["pathway_assoc"].items():
            if padmetRef and pwy_id in padmetRef.dicOfNode.keys():
                pwy_cname = padmetRef.dicOfNode[pwy_id].misc.get("COMMON-NAME",[None])[0]
                if pwy_cname:
                    line = "* [["+pwy_id+"]], "+pwy_cname+":"
            else:
                line = "* [["+pwy_id+"]]:"

            #if external link known, adding external link to the pathway
            #line: PWY_ID, common name, extern link to PWY
            if ext_link.get("pathway"):
                line += ' ['+ext_link.get("pathway")+pwy_id+' '+pwy_id+']'
            dataInArray.append(line)

            if len(total_padmet_source) > 1:
                for padmet_source, pwy_rxn in pwy_dict.items():
                    line = "** In organism: [[%s]]" %padmet_source
                    dataInArray.append(line)
                    line = "*** '''"+str(pwy_rxn["nb_found_rxn"])+"''' reactions found over '''"+str(pwy_rxn["nb_total_rxn"])+"''' reactions in the full pathway"
                    dataInArray.append(line)
            else:
                pwy_rxn = list(pwy_dict.values())[0]
                line = "** '''"+str(pwy_rxn["nb_found_rxn"])+"''' reactions found over '''"+str(pwy_rxn["nb_total_rxn"])+"''' reactions in the full pathway"
                dataInArray.append(line)

        dataInArray.append('== Reconstruction information  ==')
        if len(total_padmet_source) > 1:
            for tool, tool_dict in total_padmet_data["reconstruction"]["tool"].items():
                try:
                    tool_org_list = tool_dict[page_id]
                except KeyError:
                    tool_org_list = list()
                add_property(properties, "nb organism associated based on %s tool"%tool, [len(tool_org_list)])
            for category, category_dict in total_padmet_data["reconstruction"]["category"].items():
                try:
                    category_org_list = category_dict[page_id]
                except KeyError:
                    category_org_list = list()
                add_property(properties, "nb organism associated based on %s category"%category, [len(category_org_list)])
          
            for padmet_source, rec_data in page_dict_data["reconstruction_data"].items():
                line="* In organism: [[%s]]" %padmet_source
                dataInArray.append(line)

                for rec_dict in rec_data:
                    line = "** category: [[%s]]; source: [[%s]]; tool: [[%s]]; comment: %s" %(rec_dict["category"], rec_dict["source"], rec_dict["tool"], rec_dict["comment"])
                    dataInArray.append(line)
        else:
            rec_data = list(page_dict_data["reconstruction_data"].values())[0]
            add_property(properties, "reconstruction category", [rec_dict["category"] for rec_dict in rec_data])
            add_property(properties, "reconstruction tool", [rec_dict["tool"] for rec_dict in rec_data])
            add_property(properties, "reconstruction comment", [rec_dict["comment"] for rec_dict in rec_data])
            add_property(properties, "reconstruction source", [rec_dict["source"] for rec_dict in rec_data])
            for rec_dict in rec_data:
                line = "* category: [[%s]]; source: [[%s]]; tool: [[%s]]; comment: %s" %(rec_dict["category"], rec_dict["source"], rec_dict["tool"], rec_dict["comment"])
                dataInArray.append(line)

    elif category == "gene":
        dataInArray.append('== Organism(s) associated with this gene  ==')
        add_property(properties, "organism associated", page_dict_data["padmet_source"])
        for padmet_source in page_dict_data["padmet_source"]:
            line = "* [[%s]]" %padmet_source
            dataInArray.append(line)
        dataInArray.append("== Reaction(s) associated ==")
        add_property(properties, "nb reaction associated", [len(page_dict_data["reac_assoc"].keys())])
        for rxn_id, rxn_dict in page_dict_data["reac_assoc"].items():
            line = "* [[%s]]" %rxn_id
            dataInArray.append(line)
            if len(total_padmet_source) > 1:
                for padmet_source, link_data in rxn_dict.items():
                    line = "** In organism: [[%s]]" %padmet_source
                    dataInArray.append(line)
                    for category, category_dict in link_data.items():
                        line = "*** Category: [[%s]]" %category
                        dataInArray.append(line)
                        for source_dict in category_dict:
                            line = "**** source: [[%s]]; tool: [[%s]]; comment: %s" %(source_dict["source"], source_dict["tool"], source_dict["comment"])
                            dataInArray.append(line)
            else:
                link_data = list(rxn_dict.values())[0]
                for category, category_dict in link_data.items():
                    line = "** Category: [[%s]]" %category
                    dataInArray.append(line)
                    for source_dict in category_dict:
                        line = "*** source: [[%s]]; tool: [[%s]]; comment: %s" %(source_dict["source"], source_dict["tool"], source_dict["comment"])
                        dataInArray.append(line)

        if page_dict_data["pathway_assoc"]:
            add_property(properties, "nb pathway associated", [len(page_dict_data["pathway_assoc"].keys())])
            dataInArray.append("== Pathway(s) associated ==")
            for pwy_id, pwy_dict in page_dict_data["pathway_assoc"].items():
                line = "* [[%s]]" %pwy_id
                dataInArray.append(line)
                if len(total_padmet_source) > 1:
                    for padmet_source, pwy_rxn in pwy_dict.items():
                        line = "** In organism [[%s]]: '''%s''' reactions found over '''%s''' reactions in the full pathway" %(padmet_source, pwy_rxn["nb_found_rxn"], pwy_rxn["nb_total_rxn"])
                else:
                    pwy_rxn = list(pwy_dict.values())[0]
                    line = "** '''%s''' reactions found over '''%s''' reactions in the full pathway" %(pwy_rxn["nb_found_rxn"], pwy_rxn["nb_total_rxn"])
                dataInArray.append(line)

    elif category == "pathway":
        if len(total_padmet_source) > 1:
            dataInArray.append('== Organism(s) associated with this pathway  ==')
            add_property(properties,"nb organism associated", [len(page_dict_data["padmet_source"])])
            rate_list = [rate for rate in page_dict_data["completion_rate"].values()]
            try:
                min_completion_rate = min(rate_list)
                max_completion_rate = max(rate_list)
                avg_completion_rate = round(sum(rate_list)/len(rate_list),2)
            except TypeError:
                min_completion_rate = "n.a"
                max_completion_rate = "n.a"
                avg_completion_rate = "n.a"
            add_property(properties,"min completion rate", [min_completion_rate])
            add_property(properties,"max completion rate", [max_completion_rate])
            add_property(properties,"avg completion rate", [avg_completion_rate])
            for padmet_source in page_dict_data["padmet_source"]:
                line = "* [[%s]]" %padmet_source
                dataInArray.append(line)
        else:
            nb_reaction_found = len(page_dict_data["reac_assoc"].keys())
            completion_rate = page_dict_data["completion_rate"][total_padmet_source[0]]
            if completion_rate is None:
                completion_rate = "n.a"
            add_property(properties,"nb reaction found", [nb_reaction_found])
            add_property(properties,"completion rate", [completion_rate])


        dataInArray.append("== Reaction(s) found ==")
        for rxn_id, set_padmet_source in page_dict_data["reac_assoc"].items():
            line = "* [[%s]]" %rxn_id
            dataInArray.append(line)
            if set_padmet_source == "_1_IN_ALL_":
                line = "** Reaction found in all organisms"
                dataInArray.append(line)
            if set_padmet_source not in ["_1_IN_ALL_", "_1_IN_1_"]:
                line = "** Reaction found in %s organism(s): %s" %(len(set_padmet_source), ";".join(set_padmet_source))
                dataInArray.append(line)
        total_rxn_assoc = page_dict_data["total_reac_assoc"]
        if len(total_padmet_source) > 1:
            dataInArray.append("== Reaction(s) not found in any organism ==")
        else:
            dataInArray.append("== Reaction(s) not found ==")
        if total_rxn_assoc:
            add_property(properties,"nb total reaction", [len(total_rxn_assoc)])
            rxn_not_found = page_dict_data["total_reac_assoc"].difference(page_dict_data["reac_assoc"].keys())
            if rxn_not_found:
                for rxn_id in rxn_not_found:
                    line = "* [{0}{1} {1}]".format(ext_link.get(category), rxn_id)
                    dataInArray.append(line)
            else:
                line = "All reactions of this pathways are in present"
                dataInArray.append(line)
        else:
            add_property(properties,"nb total reaction", ["n.a"])
            line = "No padmetRef was given during wikipage creation or pathway not in metacyc, data not available"
            dataInArray.append(line)

    elif category == "metabolite":
        dataInArray.append("== Reaction(s) known to consume the compound ==")
        if page_dict_data["consumed_by"]:
            #add_property(properties, "consumed by", page_dict_data["consumed_by"].keys())
            for rxn_id, set_padmet_source in page_dict_data["consumed_by"].items():
                line = "* [[%s]]" %rxn_id
                dataInArray.append(line)
                if len(total_padmet_source) > 1:
                    for padmet_source in set_padmet_source:
                        line = "** In organism [[%s]]" %padmet_source
                        dataInArray.append(line)
                        
        dataInArray.append("== Reaction(s) known to produce the compound ==")
        if page_dict_data["produced_by"]:
            #add_property(properties, "produced by", page_dict_data["produced_by"].keys())
            for rxn_id, set_padmet_source in page_dict_data["produced_by"].items():
                line = "* [[%s]]" %rxn_id
                dataInArray.append(line)
                if len(total_padmet_source) > 1:
                    for padmet_source in set_padmet_source:
                        line = "** In organism [[%s]]" %padmet_source
                        dataInArray.append(line)

        dataInArray.append("== Reaction(s) of unknown directionality ==")
        if page_dict_data["unknown_dir_by"]:
            #add_property(properties, "associated to unknown reaction direction", page_dict_data["unknown_dir_by"].keys())
            for rxn_id, set_padmet_source in page_dict_data["unknown_dir_by"].items():
                line = "* [[%s]]" %rxn_id
                dataInArray.append(line)
                if len(total_padmet_source) > 1:
                    for padmet_source in set_padmet_source:
                        line = "** In organism [[%s]]" %padmet_source
                        dataInArray.append(line)

    elif category == "organism":
        dataInArray.append("== Summary ==")
        nb_rxn, nb_gene, nb_pwy = [len(page_dict_data[i].keys()) for i in ["reaction","gene","pathway"]]
        add_property(properties, "nb reaction associated", [nb_rxn])
        add_property(properties, "nb gene associated", [nb_gene])
        add_property(properties, "nb pathway associated", [nb_pwy])
        dataInArray.append("* %s reaction(s)" %nb_rxn)
        dataInArray.append("* %s gene(s)" %nb_gene)
        dataInArray.append("* %s pathway(s)" %nb_pwy)
        dataInArray.append("== Reaction(s) associated ==")
        for rxn_id in page_dict_data["reaction"].keys():
            line = "* [[%s]]" %rxn_id
            dataInArray.append(line)
        dataInArray.append("== Gene(s) associated ==")
        for gene_id in page_dict_data["gene"].keys():
            line = "* [[%s]]" %gene_id
            dataInArray.append(line)
        dataInArray.append("== Pathway(s) associated ==")
        for pwy_id in page_dict_data["pathway"].keys():
            line = "* [[%s]]" %pwy_id
            dataInArray.append(line)
        
    """
    elif category == "diffCondition":
        genes_up = set([rlt.id_in for rlt in padmetSpec.dicOfRelationOut[page_node.id] if rlt.type == "is_de_in" and float(rlt.misc["LOG2FC"][0]) > float(0)])
        genes_down = set([rlt.id_in for rlt in padmetSpec.dicOfRelationOut[page_node.id] if rlt.type == "is_de_in" and float(rlt.misc["LOG2FC"][0]) < float(0)])
        rxns_up = set()
        rxns_down = set()
        rxns_unknown = set()
        pwys_up = set()
        pwys_down = set()
        pwys_unknown = set()
        dataInArray.append("== %s Gene(s) up-regulated ==" %len(genes_up))
        for gene_id in genes_up:
            dataInArray.append("* [[%s]]" %gene_id)
            rxns = set([rlt.id_in for rlt in padmetSpec.dicOfRelationOut[gene_id] if rlt.type == "is_linked_to"])
            rxns_up.update(rxns)
        dataInArray.append("== %s Gene(s) down-regulated ==" %len(genes_down))
        for gene_id in genes_down:
            dataInArray.append("* [[%s]]" %gene_id)
            rxns = set([rlt.id_in for rlt in padmetSpec.dicOfRelationOut[gene_id] if rlt.type == "is_linked_to"])
            rxns_down.update(rxns)

        rxns_unknown = rxns_up.intersection(rxns_down)
        if rxns_unknown:
            rxns_up.difference_update(rxns_down)
            rxns_down.difference_update(rxns_up)

        dataInArray.append("== %s Reaction(s) exclusively linked to up-regulated gene(s) ==" %len(rxns_up))
        for rxn_id in rxns_up:
            dataInArray.append("* [[%s]]" %rxn_id)
            pwys = set([rlt.id_out for rlt in padmetSpec.dicOfRelationIn[rxn_id] if rlt.type == "is_in_pathway" and rlt.id_out in total_pwy_id])
            pwys_up.update(pwys)
        dataInArray.append("== %s Reaction(s) exclusively linked to down-regulated gene(s) ==" %len(rxns_down))
        for rxn_id in rxns_down:
            dataInArray.append("* [[%s]]" %rxn_id)
            pwys = set([rlt.id_out for rlt in padmetSpec.dicOfRelationIn[rxn_id] if rlt.type == "is_in_pathway" and rlt.id_out in total_pwy_id])
            pwys_down.update(pwys)
        dataInArray.append("== %s Reaction(s) linked to up and down regulated genes ==" %len(rxns_unknown))
        for rxn_id in rxns_unknown:
            dataInArray.append("* [[%s]]" %rxn_id)
            pwys = set([rlt.id_out for rlt in padmetSpec.dicOfRelationIn[rxn_id] if rlt.type == "is_in_pathway" and rlt.id_out in total_pwy_id])
            pwys_unknown.update(pwys)

        pwys_unknown.update(pwys_up.intersection(pwys_down))
        if pwys_unknown:
            pwys_up.difference_update(pwys_down)
            pwys_down.difference_update(pwys_up)

        dataInArray.append("== %s Pathway(s) only with reactions exclusively linked to up-regulated gene(s) ==" %len(pwys_up))
        for pwy_id in pwys_up:
            dataInArray.append("* [[%s]]" %pwy_id)
        dataInArray.append("== %s Pathway(s) only with reactions exclusively linked to down-regulated gene(s) ==" %len(pwys_down))
        for pwy_id in pwys_down:
            dataInArray.append("* [[%s]]" %pwy_id)
        dataInArray.append("== %s Pathway(s) with reactions linked to up and donw regulated genes ==" %len(pwys_unknown))
        for pwy_id in pwys_unknown:
            dataInArray.append("* [[%s]]" %pwy_id)

        add_property(properties, "gene up-regulated", genes_up)
        add_property(properties, "gene down-regulated", genes_down)
        add_property(properties, "reaction up-regulated", rxns_up)
        add_property(properties, "reaction down-regulated", rxns_down)
        add_property(properties, "reaction up-and-down-regulated", rxns_down)
        add_property(properties, "pathway up-regulated", pwys_up)
        add_property(properties, "pathway down-regulated", pwys_down)
        add_property(properties, "pathway up-and-down-regulated", pwys_unknown)
    """
    if "xref" in page_dict_data.keys():
        dataInArray.append('== External links  ==')
        for db, ids in page_dict_data["xref"].items():
            xrefLink(dataInArray, db, ids)

    title_indices = [i for i, x in enumerate(dataInArray) if str(x).startswith("==")]
    for i in range(len(title_indices)):
        try:
            if title_indices[i+1] - title_indices[i] > to_collapse_cutoff:
                dataInArray[title_indices[i]+1: title_indices[i+1]] = add_collapsible(dataInArray[title_indices[i]+1: title_indices[i+1]])
                title_indices = [i for i, x in enumerate(dataInArray) if str(x).startswith("==")]
        except IndexError:
            if len(dataInArray) - title_indices[i] > to_collapse_cutoff:
                dataInArray[title_indices[i]+1:] = add_collapsible(dataInArray[title_indices[i]+1:])
        

    dataInArray.extend(properties)
    with open(output_file,'w', encoding="utf8") as f_handler:
        for line in dataInArray:
            f_handler.write(line+"\n")
    """                
    for i in dataInArray:
        print(i)
    print("\n")
    """
def add_property(properties, prop_id, prop_values):
    """
    #TODO
    """
    prop_values = set([str(i) for i in prop_values])
    start_line = "{{#set: "+prop_id+"="
    values_part = "|".join(prop_values)
    end_line = "}}"
    toInsert = start_line + values_part + end_line
    properties.append(toInsert)

def add_collapsible(text_array, title=None):
    """
    #TODO
    """
    collapsible = ['<div class="toccolours mw-collapsible mw-collapsed" style="width:100%; overflow:auto;">']
    if title:
        collapsible.append('<div style="font-weight:bold;line-height:1.6;">%s</div>' %title)
        collapsible.append('<div class="mw-collapsible-content">')
        collapsible.extend(text_array)
        collapsible.append('</div></div>')
    else:
        collapsible.extend(text_array)
        collapsible.append('</div>')
    return collapsible        

def xrefLink(dataInArray, db, ids):
    """
    #TODO
    """
    if db == "METACYC":
        dataInArray.append("* "+db+":")
        for _id in ids:
            toInsert = "** [http://pubchem.ncbi.nlm.nih.gov/summary/summary.cgi?cid="+_id+" "+_id+"]"
            dataInArray.append(toInsert)
    elif db == "UNIPROT":
        dataInArray.append("* "+db+":")
        for _id in ids:
            toInsert = "** [http://www.uniprot.org/uniprot/"+_id+" "+_id+"]"
            dataInArray.append(toInsert)

    elif db == "KEGG" or db.startswith("LIGAND"):
        dataInArray.append("* "+db+":")
        for _id in ids:
            toInsert = "** [http://www.genome.jp/dbget-bin/www_bget?"+_id+" "+_id+"]"
            dataInArray.append(toInsert)

    elif db == "RHEA":
        dataInArray.append("* "+db+":")
        for _id in ids:
            toInsert = "** [http://www.ebi.ac.uk/rhea/reaction.xhtml?id="+_id+" "+_id+"]"
            dataInArray.append(toInsert)

    elif db == "WIKIPEDIA":
        dataInArray.append("* "+db+":")
        for _id in ids:
            toInsert = "** [http://en.wikipedia.org/wiki/"+_id+" "+_id+"]"
            dataInArray.append(toInsert)

    elif db == "CHEBI":
        dataInArray.append("* "+db+":")
        for _id in ids:
            toInsert = "** [http://www.ebi.ac.uk/chebi/searchId.do?chebiId="+_id+" "+_id+"]"
            dataInArray.append(toInsert)

    elif db == "PUBCHEM":
        dataInArray.append("* "+db+":")
        for _id in ids:
            toInsert = "** [http://pubchem.ncbi.nlm.nih.gov/summary/summary.cgi?cid="+_id+" "+_id+"]"
            dataInArray.append(toInsert)

    elif db == "ECOCYC":
        dataInArray.append("* "+db+":")
        for _id in ids:
            toInsert = "** [http://metacyc.org/ECOLI/NEW-IMAGE?object="+_id+" "+_id+"]"
            dataInArray.append(toInsert)

    elif db == "CHEMSPIDER":
        dataInArray.append("* "+db+":")
        for _id in ids:
            toInsert = "** [http://www.chemspider.com/Chemical-Structure."+_id+".html "+_id+"]"
            dataInArray.append(toInsert)

    elif db == "umbbd-compounds":
        dataInArray.append("* "+db+":")
        for _id in ids:
            toInsert = "** [http://umbbd.ethz.ch/servlets/pageservlet?ptype=c&compID="+_id+" "+_id+"]"
            dataInArray.append(toInsert)

    elif db == "ARACYC":
        dataInArray.append("* "+db+":")
        for _id in ids:
            toInsert = "** [http://metacyc.org/ARA/NEW-IMAGE?object="+_id+" "+_id+"]"
            dataInArray.append(toInsert)

    elif db == "PIR":
        dataInArray.append("* "+db+":")
        for _id in ids:
            toInsert = "** [http://pir.georgetown.edu/cgi-bin/nbrfget?uid="+_id+" "+_id+"]"
            dataInArray.append(toInsert)

    elif db == "NCI":
        dataInArray.append("* "+db+":")
        for _id in ids:
            toInsert = "** [http://cactus.nci.nih.gov/ncidb2.2/?nsc="+_id+" "+_id+"]"
            dataInArray.append(toInsert)

    elif db == "knapsacK":
        dataInArray.append("* "+db+":")
        for _id in ids:
            toInsert = "** [http://kanaya.naist.jp/knapsack_jsp/information.jsp?word="+_id+" "+_id+"]"
            dataInArray.append(toInsert)

    else:
        for _id in ids:
            toInsert = "* "+db+" : "+_id
            dataInArray.append(toInsert)


def create_navigation_page(total_padmet_data, navigation_folder, verbose=False):
    """
    #TODO
    """
    total_padmet_source = total_padmet_data["padmet_source"].keys()
    sideBarData = ["* navigation",
                   "** mainpage|mainpage-description",
                   "** Special:Ask|SMW-Ask",
                   "** workflow|workflow command history",
                   "** randompage-url|randompage",
                   "** Special:ListFiles|Files",
                   "* Metabolic network components"]

    category = "organism"
    sideBarData.append("** Category:{0}|{0}".format(category))
    fileName = os.path.join(navigation_folder, "Category:%s" %category)
    if verbose: print("Category: %s" %category)
    dataInArray = ["{{#ask: [[Category:%s]]" %category,
                   "| ?nb reaction associated",
                   "| ?nb gene associated", 
                   "| ?nb pathway associated",
                   "|sort=nb reaction associated", 
                   "|order=descending",
                   "}}"]
    with open(fileName,'w') as f:
        for line in dataInArray:
            f.write(line+"\n")

    category = "reaction"
    sideBarData.append("** Category:{0}|{0}".format(category))
    fileName = os.path.join(navigation_folder, "Category:%s" %category)
    if verbose: print("Category: %s" %category)
    
    if len(total_padmet_source) > 1:
        dataInArray = ["{{#ask: [[Category:%s]]"%category,
                       "| ?common-name",
                       "| ?ec-number",
                       "| ?nb organism associated",
                       "| ?min gene associated",
                       "| ?max gene associated",
                       "| ?avg gene associated",
                       "|sort=nb organism associated", 
                       "|order=descending"]

        dataInArray.extend(["| ?nb organism associated based on %s tool"%tool for tool in total_padmet_data["reconstruction"]["tool"].keys()]+["}}"])
    else:
        dataInArray = ["{{#ask: [[Category:%s]]"%category,
                       "| ?common-name",
                       "| ?ec-number",
                       "| ?nb gene associated",
                       "| ?nb pathway associated",
                       "| ?nb reconstruction source",
                       "| ?reconstruction category",
                       "| ?reconstruction tool",
                       "| ?reconstruction comment",
                       "}}"]
        
    with open(fileName,'w') as f:
        for line in dataInArray:
            f.write(line+"\n")

    category = "gene"
    sideBarData.append("** Category:{0}|{0}".format(category))
    fileName = os.path.join(navigation_folder, "Category:%s" %category)
    if verbose: print("Category: %s" %category)
    dataInArray = ["{{#ask: [[Category:%s]]"%category,
                   "| ?organism associated",
                   "| ?nb reaction associated",
                   "| ?nb pathway associated",
                   "|sort=nb reaction associated", 
                   "|order=descending",
                   "}}"]
    with open(fileName,'w') as f:
        for line in dataInArray:
            f.write(line+"\n")

    category = "pathway"
    sideBarData.append("** Category:{0}|{0}".format(category))
    fileName = os.path.join(navigation_folder, "Category:%s" %category)
    if verbose: print("Category: %s" %category)
    if len(total_padmet_source) > 1:
        dataInArray = ["{{#ask: [[Category:%s]]"%category,
                       "| ?common-name",
                       "| ?nb organism associated",
                       "| ?nb total reaction",
                       "| ?min completion rate",
                       "| ?max completion rate",
                       "| ?avg completion rate",
                       "|sort=avg completion rate", 
                       "|order=descending",
                       "}}"]
    else:
        dataInArray = ["{{#ask: [[Category:%s]]"%category,
                       "| ?common-name",
                       "| ?nb reaction found",
                       "| ?nb total reaction",
                       "| ?completion rate",
                       "|sort=completion rate, nb total reaction", 
                       "|order=descending",
                       "}}"]
        
    with open(fileName,'w') as f:
        for line in dataInArray:
            f.write(line+"\n")

    category = "metabolite"
    sideBarData.append("** Category:{0}|{0}".format(category))
    fileName = os.path.join(navigation_folder, "Category:%s" %category)
    if verbose: print("Category: %s" %category)
    dataInArray = ["{{#ask: [[Category:%s]]"%category,
                   "| ?common-name",
                   "}}"]
    with open(fileName,'w') as f:
        for line in dataInArray:
            f.write(line+"\n")
    """
    if all_diffCdt:
        category = "diffCondition"
        sideBarData.append("* Transcriptomic data")
        sideBarData.append("** Category:"+category+"|"+category)
        fileName = output_folder+"Category:diffCondition"
        if verbose: print("Category: %s" %category)
        dataInArray = ["{{#ask: [[Category:diffCondition]]","| ?gene up-regulated","| ?reaction up-regulated","| ?pathway up-regulated","| ?gene down-regulated","| ?reaction down-regulated","| ?pathway down-regulated","| ?reaction up-and-down-regulated","| ?pathway up-and-down-regulated""}}"]
        with open(fileName,'w') as f:
            for line in dataInArray:
                f.write(line+"\n")
    """
    sideBarData.append("* Reconstruction categories")
    all_categories = total_padmet_data["reconstruction"]["category"].keys()
    [sideBarData.append("** "+rec_category+"|"+rec_category) for rec_category in sorted(all_categories)]
    for rec_category in all_categories:
        fileName = os.path.join(navigation_folder, rec_category)
        if verbose: print("Reconstruction category: %s" %rec_category)
        if len(total_padmet_source) > 1:
            dataInArray = ["{{#ask: [[Category:reaction]]",
                           "[[nb organism associated based on %s category::>1]]"%rec_category,
                           "| ?common-name",
                           "| ?ec-number",
                           "| ?nb organism associated",
                           "| ?min gene associated",
                           "| ?max gene associated",
                           "| ?avg gene associated",
                           "| ?nb organism associated based on %s category"%rec_category, 
                           "|sort=nb organism associated based on %s category"%rec_category, 
                           "|order=descending",
                           "}}"]
        else:
            dataInArray = ["{{#ask: [[Category:reaction]] [[reconstruction category::%s]]" %rec_category,
                           "| ?common-name",
                           "| ?ec-number",
                           "| ?reconstruction tool",
                           "| ?reconstruction source",
                           "| ?reconstruction comment",
                           "| ?nb gene associated",
                           "| ?nb pathway associated",
                           "}}"]

        with open(fileName,'w') as f:
            for line in dataInArray:
                f.write(line+"\n")

    sideBarData.append("* Reconstruction tools")
    all_tools = total_padmet_data["reconstruction"]["tool"].keys()
    [sideBarData.append("** "+rec_tool+"|"+rec_tool) for rec_tool in sorted(all_tools)]
    for rec_tool in all_tools:
        fileName = os.path.join(navigation_folder, rec_tool)
        if verbose: print("Reconstruction tool: %s" %rec_tool)
        if len(total_padmet_source) > 1:
            dataInArray = ["{{#ask: [[Category:reaction]]",
                           "[[nb organism associated based on %s tool::>1]]"%rec_tool,
                           "| ?common-name",
                           "| ?ec-number",
                           "| ?nb organism associated",
                           "| ?min gene associated",
                           "| ?max gene associated",
                           "| ?avg gene associated",
                           "| ?nb organism associated based on %s tool"%rec_tool, 
                           "|sort=nb organism associated based on %s tool"%rec_tool, 
                           "|order=descending",
                           "}}"]
        else:
            dataInArray = ["{{#ask: [[Category:reaction]] [[reconstruction tool::%s]]" %rec_tool,
                           "| ?common-name",
                           "| ?ec-number",
                           "| ?reconstruction category",
                           "| ?reconstruction source",
                           "| ?reconstruction comment",
                           "| ?nb gene associated",
                           "| ?nb pathway associated",
                           "}}"]

        with open(fileName,'w') as f:
            for line in dataInArray:
                f.write(line+"\n")

    """
    sideBarData.append("* Reconstruction sources")
    all_sources = total_padmet_data["reconstruction"]["source"].keys()
    [sideBarData.append("** "+rec_source+"|"+rec_source) for rec_source in sorted(all_sources)]

    for rec_source in set(all_sources).difference(total_padmet_data["padmet_source"].keys()):
        fileName = os.path.join(navigation_folder, rec_source)
        if verbose: print("Reconstruction source: %s" %rec_source)
        dataInArray = ["{{#ask: [[Category:Reaction]] [[reconstruction source::"+rec_source+"]]","| ?COMMON NAME","| ?ec number",
                          "| ?reconstruction category","| ?reconstruction tool","| ?reconstruction source","| ?reconstruction comment","| ?gene associated","| ?in pathway","}}"]
        with open(fileName,'w') as f:
            for line in dataInArray:
                f.write(line+"\n")
    """
    if verbose: print("SideBar page")
    fileName = os.path.join(navigation_folder, "MediaWiki:Sidebar")
    with open(fileName, 'w') as f:
        for line in sideBarData:
            f.write(line+"\n")


def create_venn():
    """
    #TODO
    """
    if verbose: print("Venn Diagramm")
    #'ARGSUCCINSYN-RXN'
    categories_dict ={}
    """
    all_categories = ["orthology","annotation","gap-filling","manual","microbiont"]
    for category in all_categories:
        categories_dict[category] = set()
    """
    for rxn_id, rxn_src_dict in list(full_sources_dict.items()):
        for category in list(rxn_src_dict.keys()):
            try:
                categories_dict[category].add(rxn_id)
            except KeyError:
                categories_dict[category] = set(rxn_id)
                
    
    labels = get_labels(list(categories_dict.values()))
    fig, ax = venn4(labels, names=list(categories_dict.keys()))
    fig.savefig(wiki_folder+"files/venn.png")

def copy_io_files():
    """
    """
    #toDo in futur
        
def create_main(wiki_id):
    """
    #TODO
    """
    if verbose: print("Main page")
    ### create main page
    for line in main_template:
        main_template[main_template.index(line)] = line.replace("MODEL_ID",wiki_id)
    final_network_index = main_template.index([line for line in main_template if line.startswith("The automatic")][0])
    main_template[final_network_index] = main_template[final_network_index].replace("NB_RXN", str(len(all_rxns))).replace("NB_CPD", str(len(all_cpds))).replace("NB_PWY", str(len(all_pwys))).replace("NB_GENE", str(len(all_genes)))
    reconstruct_summary = {"ANNOTATION":0,"ORTHOLOGY":{},"MANUAL":0,"GAP-FILLING":0,"MICROBIONT":0,"HOST":0}
    for rec_node in [node for node in list(padmetSpec.dicOfNode.values()) if node.type == "reconstructionData"]:
        cat = rec_node.misc["CATEGORY"][0]
        if cat == "ORTHOLOGY":
            source = rec_node.misc["SOURCE"][0].replace("OUTPUT_PANTOGRAPH_","")
            try:
                reconstruct_summary["ORTHOLOGY"][source] += 1
            except KeyError:
                reconstruct_summary["ORTHOLOGY"][source] = 1
        else:
            reconstruct_summary[cat] += 1
    index = 1
    if reconstruct_summary["ANNOTATION"] != 0:
        main_template.insert(final_network_index+index, "* Based on annotation data:")
        index += 1
        main_template.insert(final_network_index+index, "** Tool: [http://bioinformatics.ai.sri.com/ptools/ Pathway tools]")
        index += 1
        main_template.insert(final_network_index+index, "*** Creation of a metabolic network containing "+str(reconstruct_summary["ANNOTATION"])+" reactions")
        index += 1
    if reconstruct_summary["ORTHOLOGY"] != 0:
        main_template.insert(final_network_index+index, "* Based on orthology data:")
        index += 1
        main_template.insert(final_network_index+index, "** Tool: [http://pathtastic.gforge.inria.fr Pantograph]")
        index += 1        
        for k,v in list(reconstruct_summary["ORTHOLOGY"].items()):
            main_template.insert(final_network_index+index, "*** From template ''"+k+"'' creation of a metabolic network containing: "+str(v)+" reactions")
            index += 1
    if reconstruct_summary["MANUAL"] != 0:
        main_template.insert(final_network_index+index, "* Based on expertise:")
        index += 1
        main_template.insert(final_network_index+index, "*** "+str(reconstruct_summary["MANUAL"])+" reaction(s) added")
        index += 1
    if reconstruct_summary["GAP-FILLING"] != 0:
        main_template.insert(final_network_index+index, "* Based on gap-filling:")
        index += 1
        main_template.insert(final_network_index+index, "** Tool: [https://pypi.python.org/pypi/meneco meneco]")
        index += 1
        main_template.insert(final_network_index+index, "*** "+str(reconstruct_summary["GAP-FILLING"])+" reaction(s) added")

    fileName = wiki_folder+"/navigation/Main_Page"
    with open(fileName,'w') as f:
        for line in main_template:
            f.write(line+"\n")                


def get_labels(data, fill=["number"]):
    """    
    get a dict of labels for groups in data
    example:
    In [12]: get_labels([range(10), range(5,15), range(3,8)], fill=["number"])
    Out[12]:
    {'001': '0',
    '010': '5',
    '011': '0',
    '100': '3',
    '101': '2',
    '110': '2',
    '111': '3'}
    
    Parameters
    ----------    
    data: list
        data to get label for
    fill: 
        ["number"|"logic"|"percent"]

    Returns
    -------
    dict:
        a dict of labels for different sets
    """
    N = len(data)

    sets_data = [set(data[i]) for i in range(N)]  # sets for separate groups
    s_all = set(chain(*data))                             # union of all sets

    # bin(3) --> '0b11', so bin(3).split('0b')[-1] will remove "0b"
    set_collections = {}
    for n in range(1, 2**N):
        key = bin(n).split('0b')[-1].zfill(N)
        value = s_all
        sets_for_intersection = [sets_data[i] for i in range(N) if  key[i] == '1']
        sets_for_difference = [sets_data[i] for i in range(N) if  key[i] == '0']
        for s in sets_for_intersection:
            value = value & s
        for s in sets_for_difference:
            value = value - s
        set_collections[key] = value

    labels = {k: "" for k in set_collections}
    if "logic" in fill:
        for k in set_collections:
            labels[k] = k + ": "
    if "number" in fill:
        for k in set_collections:
            labels[k] += str(len(set_collections[k]))
    if "percent" in fill:
        data_size = len(s_all)
        for k in set_collections:
            labels[k] += "(%.1f%%)" % (100.0 * len(set_collections[k]) / data_size)

    return labels

def venn4(labels, names=['A', 'B', 'C', 'D'], **options):
    """
    plots a 4-set Venn diagram

    Parameters
    ----------    
    labels: dict
        a label dict where keys are identified via binary codes ('0001', '0010', '0100', ...),
        hence a valid set could look like: {'0001': 'text 1', '0010': 'text 2', '0100': 'text 3', ...}.
        unmentioned codes are considered as ''.
    names: list
        group names

    Returns
    -------
    set
        (Figure, AxesSubplot), pyplot Figure and AxesSubplot object
    """
    colors = options.get('colors', [default_colors[i] for i in range(4)])
    figsize = options.get('figsize', (12, 12))
    dpi = options.get('dpi', 96)
    
    fig = plt.figure(0, figsize=figsize, dpi=dpi)
    ax = fig.add_subplot(111, aspect='equal')
    ax.set_axis_off()
    ax.set_ylim(bottom=0.0, top=1.0)
    ax.set_xlim(left=0.0, right=1.0)
    
    # body   
    draw_ellipse(fig, ax, 0.350, 0.400, 0.72, 0.45, 140.0, colors[0])
    draw_ellipse(fig, ax, 0.450, 0.500, 0.72, 0.45, 140.0, colors[1])
    draw_ellipse(fig, ax, 0.544, 0.500, 0.72, 0.45, 40.0, colors[2])
    draw_ellipse(fig, ax, 0.644, 0.400, 0.72, 0.45, 40.0, colors[3])
    draw_text(fig, ax, 0.85, 0.42, labels.get('0001', ''))
    draw_text(fig, ax, 0.68, 0.72, labels.get('0010', ''))
    draw_text(fig, ax, 0.77, 0.59, labels.get('0011', ''))
    draw_text(fig, ax, 0.32, 0.72, labels.get('0100', ''))
    draw_text(fig, ax, 0.71, 0.30, labels.get('0101', ''))
    draw_text(fig, ax, 0.50, 0.66, labels.get('0110', ''))
    draw_text(fig, ax, 0.65, 0.50, labels.get('0111', ''))
    draw_text(fig, ax, 0.14, 0.42, labels.get('1000', ''))
    draw_text(fig, ax, 0.50, 0.17, labels.get('1001', ''))
    draw_text(fig, ax, 0.29, 0.30, labels.get('1010', ''))
    draw_text(fig, ax, 0.39, 0.24, labels.get('1011', ''))
    draw_text(fig, ax, 0.23, 0.59, labels.get('1100', ''))
    draw_text(fig, ax, 0.61, 0.24, labels.get('1101', ''))
    draw_text(fig, ax, 0.35, 0.50, labels.get('1110', ''))
    draw_text(fig, ax, 0.50, 0.38, labels.get('1111', ''))
    
    # legend
    draw_text(fig, ax, 0.13, 0.18, names[0], colors[0])
    draw_text(fig, ax, 0.18, 0.83, names[1], colors[1])
    draw_text(fig, ax, 0.82, 0.83, names[2], colors[2])
    draw_text(fig, ax, 0.87, 0.18, names[3], colors[3])
    leg = ax.legend(names, loc='best', fancybox=True)
    leg.get_frame().set_alpha(0.5)
    
    return fig, ax

def draw_ellipse(fig, ax, x, y, w, h, a, fillcolor):
    e = patches.Ellipse(
        xy=(x, y),
        width=w,
        height=h,
        angle=a,
        color=fillcolor)
    ax.add_patch(e)

def draw_text(fig, ax, x, y, text, color=[0, 0, 0, 1]):
    ax.text(
        x, y, text,
        horizontalalignment='center',
        verticalalignment='center',
        fontsize=14,
        color=color)

default_colors = [
    # r, g, b, a
    [92, 192, 98, 0.5],
    [90, 155, 212, 0.5],
    [246, 236, 86, 0.6],
    [241, 90, 96, 0.4],
    [255, 117, 0, 0.3],
    [82, 82, 190, 0.2],
]
default_colors = [
    [i[0] / 255.0, i[1] / 255.0, i[2] / 255.0, i[3]]
    for i in default_colors
]


main_template = ["== MODEL_IDGEM description ==",
     "== Automatic reconstruction with [http://aureme.genouest.org AuReMe] ==",
     "Model summary: [[MEDIA:summary.txt|summary]]",
     "",
     "Download '''AuReMe''' Input/Output [LINK OR MEDIA data]",
     "",
     "The automatic reconstruction of ''MODEL_NAME'' results to a Genome scale [[MEDIA:model.xml|Model]] containing NB_RXN reactions, NB_CPD metabolites, NB_GENE genes and NB_PWY pathways. This GeM was obtained based on the following sources:",
     "",
     "[[FILE:venn.png|frameless|border]]",
     "",
     "== Collaborative curation == ",
     "* Suggest reactions to add or remove:",
     "** Download this [[MEDIA:Add_delete_reaction.csv|form]]",
     "* Suggest new reactions to create and add:",
     "** Download this [[MEDIA:Reaction_creator.csv|form]]",
     "* '''Follow the examples given in the form(s) to correctly share your suggestions'''",
     "* Send the filled form(s) to: CONTACT_MAIL"]

def create_log_page(log_file, output_folder):
    """
    #TODO
    """
    cmd_regex = '--cmd=\"(.*)\"'
    fileName = output_folder+"workflow"
    log_page = ["=Workflow command history=","","==Command sequence=="]
    with open(log_file, 'r') as f:
        log_data = [line for line in f.read().splitlines() if not line.startswith("#")]
    for cmd_line in log_data:
        re_result = re.search(cmd_regex, cmd_line)
        if re_result:
            full_cmd = re_result.groups(1)[0]
            cmd = full_cmd.split(" ")[0]
            cmd_label,desc = get_cmd_label(cmd)
            if cmd_label and desc:
                if cmd == "curation":
                    curation_file = re.search('DATA=(.*\.\w+)\s?',full_cmd).groups(1)[0]
                    desc = desc.replace('FORM_FILE_NAME', curation_file)
                log_page.append("* '''%s''':" %cmd_label)
                log_page.append("''%s''" %desc)
    log_page.extend(["==Downloads==","You can download the [[MEDIA:log.txt|command log file here]]"])
    #todo downalosalsjka
    with open(fileName, 'w') as f:
        for line in log_page:
            f.write(line+"\n")
        
                 

def get_cmd_label(cmd):
    """
    #TODO
    """
    cmd_label_dict = {'getdb':{'CMD_LABEL':'Get database','DESC':"Display the available reference databases"},
                      'check_input':{'CMD_LABEL':'Check input','DESC':"Check the validity, consistency and presence of input files"},
                      'check_studied_organism_input':{'CMD_LABEL':'Check studied organism input','DESC':"Check if FAA or GBK was given for the studied organism"},
                      'check_model_organism_input':{'CMD_LABEL':'Check model organism input','DESC':"Check (if existing) each folder in orthology based reconstruction"},
                      'sbml_validity':{'CMD_LABEL':"Check SBML validity", 'DESC':"Check the SBML validity of $(METABOLIC_MODEL)."},
                      'faa_validity':{'CMD_LABEL':"Check Fasta validity", 'DESC':"Check if genes IDs in the model metabolic network are the same than in the Fasta file.<BR>If the rate of validated genes IDs is lower than the CUTOFF (see config.txt), raise an error and break the workflow."},
                      'check_gap_filling_input':{'CMD_LABEL':"Check inputs for gap-filling", 'DESC':"Check the seeds, targets and artefacts files."},
                      'curation':{'CMD_LABEL':"Manual curation",'DESC':"Apply the curation described in the form file FORM_FILE_NAME."},
                      'annotation_based':{'CMD_LABEL':"Annotation based reconstruction", 'DESC':"Extract network data from Pathway Tools annotation output."},
                      'pathwaytools':{'CMD_LABEL':"Pathway Tools", 'DESC':"For each folder in /annotation_based_reconstruction, check or generate from pgdb files (.dat) the SBML file."},
                      'orthology_based':{'CMD_LABEL':"Orthology based reconstruction", 'DESC':"Run the orthology based reconstruction."},
                      'inparanoid':{'CMD_LABEL':"Inparanoid", 'DESC':"Run Inparanoid : search for orthologs."},
                      'OMCL':{'CMD_LABEL':"OrthoMCL", 'DESC':"Run OrthoMCL : search for orthologs."},
                      'mp_pantograph' :{'CMD_LABEL':"Pantograph multiprocessed", 'DESC':"Run Pantograph : merges OrthoMCL and inparanoid results.<BR>Quickened by mulitprocessing."},
                      'pantograph':{'CMD_LABEL':"Pantograph", 'DESC':"Run Pantograph : merges OrthoMCL and inparanoid results."},
                      'draft':{'CMD_LABEL':"Create draft network", 'DESC':"Merges all available networks from the /networks directory into one metabolic network.<BR>Merge all data on the studied species."},
                      'gap_filling':{'CMD_LABEL':"Run gap-filling", 'DESC':"Calculate the gap-filling solution and generate the metabolic network, completed with the gap-filling solution."},
                      'gap_filling_solution':{'CMD_LABEL':"Calculate gap-filling solution", 'DESC':"Only calculate the gap-filling solution."},
                      'meneco':{'CMD_LABEL':"Meneco", 'DESC':"Run Meneco : a gap-filling reconstruction method."},
                      'final':{'CMD_LABEL':"Final network", 'DESC':"Generate the final metabolic network, once applyed all the reconstruction methods."},
                      'gbk_to_faa':{'CMD_LABEL':"GBK to Fasta", 'DESC':"Export a GeneBank (.gbk) file in Fasta (faa) format."},
                      'pgdb_to_padmet':{'CMD_LABEL':"PGDB to PADMet", 'DESC':"Export a PGDB (.dat Pathway Tools files) in PADMet (.padmet) format."},
                      'padmet_to_sbml':{'CMD_LABEL':"PADMet to SBML", 'DESC':"Export a PADMet (.padmet) file in the SBML format."},
                      'compounds_to_sbml':{'CMD_LABEL':"Compounds to SBML", 'DESC':"Export a list of compounds (.txt) in the SBML format."},
                      'sbml_mapping':{'CMD_LABEL':"SBML mapping", 'DESC':"Map an SBML file (all the entities IDs) to a reference database (specified in config.txt)."},
                      'get_medium':{'CMD_LABEL':"Get medium", 'DESC':"Display the studied species specified growth medium."},
                      'set_medium':{'CMD_LABEL':"Set medium", 'DESC':"Set the growth medium for the studied species."},
                      'del_medium':{'CMD_LABEL':"Delete medium", 'DESC':"Delete the growth medium for the studied species."},
                      'get_compart':{'CMD_LABEL':"Get compartments", 'DESC':"Display all the compartments of the metabolic network."},
                      'del_compart':{'CMD_LABEL':"Delete compartment", 'DESC':"Remove a compartment from the metabolic network."},
                      'change_compart':{'CMD_LABEL':"Change compartment", 'DESC':"Modify a compartment in the metabolic network."},
                      'report':{'CMD_LABEL':"Report", 'DESC':"Generate reports on the metabolic network reconstruction."},
                      'set_fba':{'CMD_LABEL':"Set FBA", 'DESC':"Set the biomass reaction to run flux balance analysis on the network."},
                      'summary':{'CMD_LABEL':"Test FBA", 'DESC':"Run flux balance analysis on the network."},
                      'menecheck':{'CMD_LABEL':"Menecheck", 'DESC':"Run topological analysis on the network."},
                      'shogen':{'CMD_LABEL':"Shogen", 'DESC':"Run Shogen : find shortest genome segments that regulate metabolic pathways."},
                      'wiki_pages':{'CMD_LABEL':"Create Wiki pages", 'DESC':"Create Wiki pages to display the metabolic network."},
                      'wiki_run':{'CMD_LABEL':"Run Wiki", 'DESC':"Create a Docker container for the Wiki."},
                      'wiki_init':{'CMD_LABEL':"Wiki initialization", 'DESC':"Send data on the metabolic network to the Docker container to fill in the Wiki."},
                      'send_all_page':{'CMD_LABEL':"Send all pages to Wiki", 'DESC':"Send all the generated pages on the metabolic network to the Wiki."},
                      'tsv':{'CMD_LABEL':"PADMet to tsv",'DESC':"Convert a PADMet (.padmet) file to a tsv files (for Askomics)."}
                      }
    current_cmd_dict = cmd_label_dict.get(cmd)
    if current_cmd_dict:
        cmd_label = current_cmd_dict["CMD_LABEL"]
        desc = current_cmd_dict["DESC"]
    else:
        cmd_label, desc = None, None
    return(cmd_label,desc)



if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
"""
Description:
    From a file containing a list of reaction, return the pathways where these reactions 
    are involved.
    ex: if rxn-a in pwy-x => return, pwy-x; all rxn ids in pwy-x; all rxn ids in pwy-x FROM the list; ratio

::
    
    usage:
        get_pwy_from_rxn.py --reaction_file=FILE --padmetRef=FILE  --output=FILE
    
    options:
        -h --help     Show help.
        --reaction_file=FILE    pathname of the file containing the reactions id, 1/line 
        --padmetRef=FILE    pathname of the padmet representing the database.
        --output=FILE    pathname of the file with line = pathway id, all reactions id, reactions ids from reaction file, ratio. sep = "\t"
"""
from padmet.classes import PadmetSpec
import docopt

def main():
    args = docopt.docopt(__doc__)
    reaction_file = args["--reaction_file"]
    with open(reaction_file, 'r') as f:
        reactions = set(f.read().splitlines())
    padmet_file = args["--padmetRef"]
    output = args["--output"]
    padmet = PadmetSpec(padmet_file)
    dict_pwy = extract_pwys(padmet, reactions)
    dict_pwys_to_file(dict_pwy, output)

def extract_pwys(padmet, reactions):
    """
    #extract from padmet pathways containing 1-n reactions from a set of reactions 'reactions'
    Return a dict of data.
    dict, k=pathway_id, v=dict: k in [total_rxn, rxn_from_list, ratio
    ex: {pwy-x:{'total_rxn':[a,b,c], rxn_from_list:[a], ratio:1/3}}


    Parameters
    ----------
    padmet: padmet.classes.PadmetSpec
        padmet to udpate
    reactions: set
        set of reactions to match with pathways
        
    Returns
    -------
    dict:
        dict, k=pathway_id, v=dict: k in [total_rxn, rxn_from_list, ratio
        ex: {pwy-x:{'total_rxn':[a,b,c], rxn_from_list:[a], ratio:1/3}}
    """
    dict_pwy = {}
    pwys_id = [node.id for node in padmet.dicOfNode.values() if node.type == "pathway"]
    for pwy_id in pwys_id:
        rxn_in_pwy = set([rlt.id_in for rlt in padmet.dicOfRelationOut[pwy_id] if rlt.type == "is_in_pathway" and padmet.dicOfNode[rlt.id_in].type == "reaction"])
        rxn_from_list = reactions.intersection(rxn_in_pwy)
        if rxn_from_list:
            ratio = round(float(len(rxn_from_list))/float(len(rxn_in_pwy)),2)
            dict_pwy[pwy_id] = {'total_rxn': rxn_in_pwy, 'rxn_from_list': rxn_from_list, 'ratio': ratio}
    return dict_pwy
    
def dict_pwys_to_file(dict_pwy, output):
    """
    Create csv file from dict_pwy. dict_pwy is obtained with extract_pwys()

    Parameters
    ----------
    dict_pwy: dict
        dict, k=pathway_id, v=dict: k in [total_rxn, rxn_from_list, ratio
        ex: {pwy-x:{'total_rxn':[a,b,c], rxn_from_list:[a], ratio:1/3}}
    output: str
        path to output file

    """
    with open(output, 'w') as f:
        header = ["pathway_id","total_rxn","rxn_from_list","ratio"]
        f.write("\t".join(header)+"\n")
        for pwy_id, pwy_data in dict_pwy.items():
            total_rxn = pwy_data['total_rxn']
            rxn_from_list = pwy_data['rxn_from_list']
            ratio = pwy_data['ratio']

            line = [pwy_id, ";".join(total_rxn), ";".join(rxn_from_list),str(ratio)]
            f.write("\t".join(line)+"\n")

    
if __name__ == "__main__":
    main()    



# -*- coding: utf-8 -*-
"""
Description:
    From a list of genes, get from the linked reactions the list of products.

    R1 is linked to G1, R1 produces M1 and M2.  output: M1,M2. Takes into account reversibility

::

    usage:
        gene_to_targets.py --padmetSpec=FILE --genes=FILE --output=FILE [-v]
    
    option:
        -h --help     Show help
        --padmetSpec=FILE    path to the padmet file
        --genes=FILE   path to the file containing gene ids, one id by line
        --output=FILE    path to the output file containing all tagerts which can by produced by all reactions associated to the given genes
        -v   print info
"""
from padmet.classes import PadmetSpec
import docopt

def main():
    #recovering args
    args = docopt.docopt(__doc__)
    padmet = PadmetSpec(args["--padmetSpec"])
    genes_file = args["--genes"]
    output = args["--output"]
    verbose = args["-v"]
    gene_to_targets(padmet, genes_file, output, verbose)
    
def gene_to_targets(padmet, genes_file, output, verbose=False):
    """
    From a list of genes, get from the linked reactions the list of products.
    R1 is linked to G1, R1 produces M1 and M2.  output: M1,M2. Takes into account reversibility

    Parameters
    ----------
    padmet: padmet.classes.PadmetSpec
        padmet to explore
    genes_file: str
        path of genes file, 1 gene id by line
    output: str
        pathname of the output file
    verbose: bool
        if True print information
    """
    
    with open(genes_file,'r') as f:
        all_genes = f.read().splitlines()
        nb_genes = len(all_genes)
    all_targets = set()
    count = 0
    for gene_id in all_genes:
        count += 1
        try:
            #check if gene id in padmet
            padmet.dicOfNode[gene_id]
            if verbose: print("%s/%s %s" % (count, nb_genes, gene_id))
        except KeyError:
            if verbose: print("Gene %s not found in padmetSpec" %gene_id)
            continue

            #get all reactions linked to the gene
        try:
            reactions_linked = [padmet.dicOfNode[rlt.id_in] 
            for rlt in padmet.dicOfRelationOut[gene_id]
            if rlt.type == "is_linked_to"]
        except KeyError:
            if verbose: print("the gene %s is not linked to any reaction" %gene_id)
            continue
        if verbose: print("\t%s reactions linked" %len(reactions_linked))

        for reaction_node in reactions_linked:
            reaction_node_id = reaction_node.id
            if verbose: print("\t\t"+reaction_node_id)
            #all products
            products = set([rlt.id_out 
            for rlt in padmet.dicOfRelationIn.get(reaction_node_id, None)
            if rlt.type == "produces"])
            #add the products as targets
            all_targets = all_targets.union(products)
            #if reaction is reversible (reversible or UNKNOWN direction) add reactants
            if reaction_node.misc["DIRECTION"][0] != "LEFT-TO-RIGHT":
                #all reactants
                reactants = set([rlt.id_out 
                for rlt in padmet.dicOfRelationIn.get(reaction_node_id, None)
                if rlt.type == "consumes"])
                all_targets = all_targets.union(reactants)
    
    if len(all_targets) != 0:
        with open(output, 'w') as f:
            f.write("\n".join(all_targets))
        
                
if __name__ == "__main__":
    main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Description:
    Create sbml from sbml. Use it to change sbml level.

::
    usage:
        sbml_to_sbml.py --input=FILE/FOLDER --output=FILE/FOLDER --new_sbml_lvl=STR [--cpu=INT]
    
    option:
        -h --help    Show help.
        --input=FILE    path of the sbml file/folder to convert into sbml
        --output=FILE    path of the sbml file/folder to generate.
        --new_sbml_lvl=STR    level of the new sbml.
        --cpu=FILE    number of cpu used [default: 1].
        -v   print info.
"""

import docopt
import os
import sys

from datetime import datetime
from multiprocessing import Pool
from padmet.classes import PadmetSpec
from padmet_utils.connection.sbmlGenerator import padmet_to_sbml

def main():
    args = docopt.docopt(__doc__)
    input_sbml = args["--input"]
    output_sbml = args["--output"]
    new_sbml_level = args["--new_sbml_lvl"]
    cpu = args['--cpu']

    from_sbml_to_sbml(input_sbml, output_sbml, new_sbml_level, cpu)


def run_sbml_to_sbml(multiprocess_data):
    """Turn sbml to sbml.

    Parameters
    ----------
    multiprocess_data: dict
        pathname to species sbml file, pathname to output sbml file, new sbml level

    Returns
    -------
    bool:
        True if sbml file exists
    """
    padmet = sbml_to_padmet(sbml=multiprocess_data['sbml_file'], db=None, version=None, source_tool=None,
                            source_category=None, source_id=None, mapping=None, verbose=None)
    padmet_to_sbml(padmet, multiprocess_data['sbml_output_file'], sbml_lvl=multiprocess_data['new_sbml_level'], verbose=False)

    if multiprocess_data['sbml_output_file'] and not os.access(multiprocess_data['sbml_output_file'], os.W_OK):
        try:
            open(multiprocess_data['sbml_output_file'], 'w').close()
            os.unlink(multiprocess_data['sbml_output_file'])
            return True
        except OSError:
            return False
    else:  # path is accessible
        return True

def sbml_to_padmet(sbml, db, version, source_tool, source_category, source_id, mapping, verbose):
    db='NA'
    version='NA'
    now = datetime.now()
    today_date = now.strftime("%Y-%m-%d")

    if os.path.isdir(sbml):
        sbml_type = "dir"
    elif os.path.isfile(sbml):
        sbml_type = "file"
    else:
        raise TypeError("%s is not a dir or a file" %(sbml))
    padmet_to_update = PadmetSpec()

    POLICY_IN_ARRAY = [['class','is_a_class','class'], ['class','has_name','name'], ['class','has_xref','xref'], ['class','has_suppData','suppData'],
                    ['compound','is_a_class','class'], ['compound','has_name','name'], ['compound','has_xref','xref'], ['compound','has_suppData','suppData'],
                    ['gene','is_a_class','class'], ['gene','has_name','name'], ['gene','has_xref','xref'], ['gene','has_suppData','suppData'], ['gene','codes_for','protein'],
                    ['pathway','is_a_class','class'], ['pathway','has_name','name'], ['pathway','has_xref','xref'], ['pathway','is_in_pathway','pathway'],
                    ['protein','is_a_class','class'], ['protein','has_name','name'], ['protein','has_xref','xref'], ['protein','has_suppData','suppData'], ['protein','catalyses','reaction'],
                    ['protein','is_in_species','class'],
                    ['reaction','is_a_class','class'], ['reaction','has_name','name'], ['reaction','has_xref','xref'], ['reaction','has_suppData','suppData'], ['reaction','has_reconstructionData','reconstructionData'], ['reaction','is_in_pathway','pathway'],
                    ['reaction','consumes','class','STOICHIOMETRY','X','COMPARTMENT','Y'], ['reaction','produces','class','STOICHIOMETRY','X','COMPARTMENT','Y'],
                    ['reaction','consumes','compound','STOICHIOMETRY','X','COMPARTMENT','Y'], ['reaction','produces','compound','STOICHIOMETRY','X','COMPARTMENT','Y'],
                    ['reaction','consumes','protein','STOICHIOMETRY','X','COMPARTMENT','Y'], ['reaction','produces','protein','STOICHIOMETRY','X','COMPARTMENT','Y'],
                    ['reaction','is_linked_to','gene','SOURCE:ASSIGNMENT','X:Y']]
    dbNotes = {"PADMET":{"creation":today_date,"version":"2.6"},"DB_info":{"DB":db,"version":version}}
    padmet_to_update.setInfo(dbNotes)
    padmet_to_update.setPolicy(POLICY_IN_ARRAY)

    #if sbml is a directory, recover all file path in a list. if no => only one file: create a list with only this file
    sbml_mapping_dict = {}
    if sbml_type == "dir":
        path = sbml
        if not path.endswith("/"):
            path += "/"
        all_files = [i for i in next(os.walk(path))[2] if not i.startswith(".~lock")]
        for sbml_file in [i for i in all_files if i.endswith(".sbml") or i.endswith(".xml")]:
            mapping_file = os.path.splitext(sbml_file)[0] + "_dict.csv"
            if mapping_file not in all_files:
                mapping_file = None
            else:
                mapping_file = path+mapping_file
            sbml_file = path+sbml_file
            sbml_mapping_dict[sbml_file] = mapping_file
    else:
        sbml_mapping_dict[sbml] = mapping

    for sbml_file, mapping_file in list(sbml_mapping_dict.items()):
        if mapping_file:
            force = False
        else:
            force = True

        padmet_to_update.updateFromSbml(sbml_file, verbose = verbose, source_category = source_category, source_id = source_id,
                                        source_tool = source_tool, mapping_file = mapping_file, force = force )

    if len(list(sbml_mapping_dict.keys())) == 0:
        if verbose: print("No sbml found in %s" %sbml)

    return padmet_to_update

def from_sbml_to_sbml(input_sbml, output_sbml, new_sbml_level, cpu=1):
    """Turn sbml to sbml.

    Parameters
    ----------
    input_sbml: str
        pathname to species sbml file/folder
    output_sbml: str
        pathname to output sbml file/folder
    new_sbml_level: int
        new sbml level
    cpu: int
        number of cpu

    Returns
    -------
    str:
        pathname to output sbml file/folder
    """
    try:
        new_sbml_level = int(new_sbml_level)
    except:
        print('SBML level must be an int.')

    if new_sbml_level not in [2, 3]:
        sys.exit('New SBML level must be 2 or 3.')

    sbml_to_sbml_pool = Pool(processes=int(cpu))

    multiprocess_datas = []
    if os.path.isdir(input_sbml):
        if not os.path.isdir(output_sbml):
            try:
                os.makedirs(output_sbml)
            except OSError:
                print('Impossible to create output folder.')
        for sbml_file in os.listdir(input_sbml):
            multiprocess_data = {}
            multiprocess_data['sbml_file'] = input_sbml + '/' + sbml_file
            multiprocess_data['sbml_output_file'] = output_sbml + '/' + sbml_file
            multiprocess_data['new_sbml_level'] = new_sbml_level
            multiprocess_datas.append(multiprocess_data)

    elif os.path.isfile(input_sbml):
        multiprocess_data = {}
        multiprocess_data['sbml_file'] = input_sbml
        multiprocess_data['sbml_output_file'] = output_sbml
        multiprocess_data['new_sbml_level'] = new_sbml_level
        multiprocess_datas.append(multiprocess_data)

    sbml_checks = sbml_to_sbml_pool.map(run_sbml_to_sbml, multiprocess_datas)

    sbml_to_sbml_pool.close()
    sbml_to_sbml_pool.join()

    if all(sbml_checks):
        return output_sbml

    else:
        print("Error during sbml creation.")

if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
"""
Description:
    compare reactions in sbml and padmet file

::
    
    usage:
        compare_sbml_padmet.py --padmet=FILE --sbml=FILE
    
    option:
        -h --help    Show help.
        --padmet=FILE    path of the padmet file
        --sbml=FILE    path of the sbml file
"""
from padmet.classes import PadmetSpec
import padmet.utils.sbmlPlugin as sp
import libsbml
import docopt

def main():
    args = docopt.docopt(__doc__)
    sbml_file = args["--sbml"]
    reader = libsbml.SBMLReader()
    sbml_document = reader.readSBML(sbml_file)
    for i in range(sbml_document.getNumErrors()):
        print(sbml_document.getError(i).getMessage())

    padmet = PadmetSpec(args["--padmet"])
    compare_sbml_padmet(sbml_document, padmet)
    
def compare_sbml_padmet(sbml_document, padmet):
    """
    compare reactions ids in sbml vs padmet, return nb of reactions in both
    and reactions id not in sbml or not in padmet

    Parameters
    ----------
    padmet: padmet.classes.PadmetSpec
        padmet to udpate
    sbml_file: libsbml.document
        sbml document
    """
    sbml_model = sbml_document.getModel()

    sbml_listOfReactions = set([sp.convert_from_coded_id(r.getId())[0] for r in sbml_model.getListOfReactions()])
    padmet_reaction = set([node.id for node in padmet.dicOfNode.values() if node.type == "reaction"])
    diff = sbml_listOfReactions.difference(padmet_reaction)
    diff_inv = padmet_reaction.difference(sbml_listOfReactions)
    
    print("%s reaction in sbml" %len(sbml_listOfReactions))
    print("%s reaction in padmet" %len(padmet_reaction))
    print("%s reaction in sbml not in padmet" %len(diff))
    for i in diff:
        print("\t%s" %i)
    print("%s reaction in padmet not in sbml" %len(diff_inv))
    for i in diff_inv:
        print("\t%s" %i)

if __name__ == "__main__":
    main()


#!/usr/bin/env python3.7
# -*- coding: utf-8 -*-
"""
usage:
    prot2genome --query_faa=FILE --query_ids=FILE/STR --subject_gbk=FILE --subject_fna=FILE --subject_faa=FILE --output_folder=FILE [--cpu=INT] [blastp] [tblastn] [debug]
    prot2genome --query_faa=FILE --query_ids=FILE/STR --subject_gbk=FILE --subject_fna=FILE --subject_faa=FILE --output_folder=FILE --exonerate=PATH  [--cpu=INT] [blastp] [tblastn] [debug]

options:
    --query_faa=FILE #TODO. 
    --query_ids=FILE/STR #TODO. 
    --subject_gbk=FILE #TODO. 
    --subject_fna=FILE #TODO. 
    --subject_faa=FILE #TODO. 
    --output_folder=FILE #TODO. 
    --cpu=INT     Number of cpu to use for the multiprocessing (if none use 1 cpu). [default: 1]
    --exonerate=PATH #TODO. 
    blastp #TODO. 
    tblastn #TODO. 
    debug #TODO. 
    
"""

from Bio.Blast.Applications import NcbiblastpCommandline, NcbitblastnCommandline
from Bio import SearchIO
from Bio import SeqIO
import os
import csv
import subprocess
from multiprocessing import Pool
import docopt

#Utilise gbk to faa, ajouter une option pour faire un fichier fasta une sequence.
#pident a ajouter
#TODO:
#Seq to gbk, match to faa


def main():
    """
    """
    global exonerate_path
    args = docopt.docopt(__doc__)
    query_faa = args["--query_faa"]
    query_ids = args["--query_ids"]
    subject_gbk = args["--subject_gbk"]
    subject_fna = args["--subject_fna"]
    subject_faa = args["--subject_faa"]
    output_folder = args["--output_folder"]
    cpu =  int(args["--cpu"])
    exonerate_path = args["--exonerate"]
    blastp = args["blastp"]
    tblastn = args["tblastn"]
    debug = args["debug"]

    if exonerate_path:
        subprocess.run([exonerate_path], shell = True)
        exonerate = True

    #if query_ids is a file, extract all query_ids (1 by line)
    #else, query_ids must represent query_ids sep by ';'
    if os.path.isfile(query_ids):
        with open(query_ids, 'r') as f:
            all_query_seq_ids = f.read().splitlines()
    else:
        all_query_seq_ids = query_ids.split(";")

    print("%s query ids to search" %len(all_query_seq_ids))
    #If GBK given and no fna, create fna from genbank
    if not os.path.exists(subject_fna):
        SeqIO.convert(subject_gbk, "genbank", subject_fna, "fasta")
    #Do the same for gbk to fasta ? (must be with padmet-utils for isoforms)
    #TODO

    #output file: header: see analysis_header from analysisOutput function
    analysis_output = os.path.join(output_folder, "blast_analysis.csv")

    pool = Pool(cpu)
    #list of dict to give to dictWritter
    all_analysis_result = []
    #Create a list of dict, each dict is the arg given to runAllAnalysis fct
    all_dict_args = []
    for query_seq_id in all_query_seq_ids:
        dict_args = {"query_seq_id": query_seq_id, "query_faa": query_faa, "subject_faa": subject_faa, "subject_fna": subject_fna, "output_folder": output_folder, "blastp": blastp, "tblastn": tblastn, "exonerate": exonerate, "debug": debug}
        all_dict_args.append(dict_args)
    #Run runAllAnalysis in multiproccess.
    mp_results = pool.map(runAllAnalysis, all_dict_args)
    for _list in mp_results:
        all_analysis_result += _list
    print("Creating output analysis: %s" %analysis_output)
    #Create output file
    analysisOutput(all_analysis_result, analysis_output)


def runBlastp(query_seq_faa, subject_faa, header=["sseqid", "evalue", "bitscore"], debug=False):
    """
    """
    print("\tRunning Blastp %s vs %s" %(os.path.basename(query_seq_faa), os.path.basename(subject_faa)))
    outfmt_arg = '"%s %s"'%(6, " ".join(header))
    output = NcbiblastpCommandline(query=query_seq_faa, subject=subject_faa, evalue=1e-10, outfmt=outfmt_arg)()[0]
    output = [line.split("\t") for line in output.splitlines()]
    blastp_result = {}
    count = 0
    for line in output:
        count += 1
        blastp_result[count] = {}
        for (index, h) in enumerate(header):
            blastp_result[count]["blastp_"+h] = line[index]
    try:
        max_bitscore = max([float(hsp["blastp_bitscore"]) for hsp in blastp_result.values()])
    except ValueError:
        max_bitscore = "N.A"
    if debug:
        print("\t\tAll hsp from blastp:")
        if blastp_result.values():
            for hsp in blastp_result.values():
                print(hsp)
            print("\t\tMax biscore: %s"%max_bitscore)
        else:
            print("\t\tNo HPS")
    try:
        result = [hsp for hsp in blastp_result.values() if float(hsp["blastp_bitscore"]) == max_bitscore][0]
    except IndexError:
        result = {}
    return result

def runTblastn(query_seq_faa, subject_fna, header=["sseqid", "evalue", "bitscore"], debug=False):
    """
    """    
    print("\tRunning tBlastn %s vs %s" %(os.path.basename(query_seq_faa), os.path.basename(subject_fna)))
    outfmt_arg = '"%s %s"'%(6, " ".join(header))
    output = NcbitblastnCommandline(query=query_seq_faa, subject=subject_fna, evalue=1e-10, outfmt=outfmt_arg)()[0]
    output = [line.split("\t") for line in output.splitlines()]
    tblastn_result = {}
    count = 0
    for line in output:
        count += 1
        tblastn_result[count] = {}
        for (index, h) in enumerate(header):
            tblastn_result[count]["tblastn_"+h] = line[index]
    try:
        max_bitscore = max([float(hsp["tblastn_bitscore"]) for hsp in tblastn_result.values()])
    except ValueError:
        max_bitscore = "N.A"
    if debug:
        print("\t\tAll hsp from tblastn:")
        for hsp in tblastn_result.values():
            print(hsp)
        print("\t\tMax biscore: %s"%max_bitscore)
    try:           
        result = [hsp for hsp in tblastn_result.values() if float(hsp["tblastn_bitscore"]) == max_bitscore][0]
    except IndexError:
        result = {}
    return result

def runExonerate(query_seq_faa, sseq_seq_faa, output, debug=False):
    """
    """
    print("\tRunning Exonerate %s vs %s" %(os.path.basename(query_seq_faa), os.path.basename(sseq_seq_faa)))
    exonerate_result = {}
    cmd_args = '{0} --model protein2genome {1} {2} --score 500 --showquerygff True  \
    --ryo ">%qi length=%ql alnlen=%qal\n>%ti length=%tl alnlen=%tal\n" >> {3}'.format(exonerate_path, query_seq_faa, sseq_seq_faa, output)
    subprocess.run([cmd_args], shell = True)
    try:
        exonerate_raw_output = list(SearchIO.parse(output, 'exonerate-text'))
       
        best_hsp = exonerate_raw_output[0][0][0]
        for qresult in exonerate_raw_output:
            for hit in qresult:
                for hsp in hit:
                    if hsp.score > best_hsp.score:
                        best_hsp = hsp
        if debug:
            print("\t\tAll hsp from exonerate:")
            for qresult in exonerate_raw_output:
                for hit in qresult:
                    for hsp in hit:
                        print(hsp)
            print("\t\tMax score: %s"%best_hsp.score)
        exonerate_result = {"exonerate_score": best_hsp.score, "exonerate_hit_range": best_hsp.hit_range}
    except RuntimeError:
        if debug:
            print("\t\tNo HSP")
    return exonerate_result


def runAllAnalysis(dict_args):
    """
    """
    query_seq_id = dict_args["query_seq_id"]
    query_faa = dict_args["query_faa"]
    subject_faa = dict_args["subject_faa"]
    subject_fna = dict_args["subject_fna"]
    output_folder = dict_args["output_folder"]
    blastp = dict_args["blastp"]
    tblastn = dict_args["tblastn"]
    exonerate = dict_args["exonerate"]
    debug = dict_args["debug"]
        
    
    with open(query_faa, "r") as faa:
        query_seqs = [seq_record for seq_record in SeqIO.parse(faa, "fasta") if seq_record.id.startswith(query_seq_id+"_isoform") or seq_record.id == query_seq_id]
    if len(query_seqs) > 1:
        print("/!\ Isoforms found for %s: %s"%(query_seq_id, [i.name for i in query_seqs]))
    analysis_result = list()
    for query_seq in query_seqs:
        query_seq_id = query_seq.id
        query_seq_faa = os.path.join(output_folder,query_seq_id+".faa")
        if not os.path.exists(query_seq_faa):
            SeqIO.write(query_seq, query_seq_faa, "fasta")
        current_result ={"query_seq_id": query_seq_id,}
    
        # Run BLASTP and parse the output
        if blastp:
            blastp_result = runBlastp(query_seq_faa, subject_faa, debug=debug)
            if blastp_result:
                current_result.update(blastp_result)
        # Run TBLASTN and parse the output
        if tblastn:
            tblastn_result = runTblastn(query_seq_faa, subject_fna, debug=debug)
            if tblastn_result:
                current_result.update(tblastn_result)
        
        # Run exonerate, parse output
        if exonerate:
            if not tblastn:
                IOError("must run tblastn to be able to run exonerate")
            _tblastn_hit = True
            try:
                exonerate_target_id = current_result["tblastn_sseqid"]
            except KeyError:
                _tblastn_hit = False
                if debug:
                    print("No hit from tblastn, can't run exonerate")
            if _tblastn_hit:
                sseq_seq_faa = os.path.join(output_folder,exonerate_target_id+".fna")
                if not os.path.exists(sseq_seq_faa):
                    with open(subject_fna, "r") as fna:
                        sseq_seq = [seq_record for seq_record in SeqIO.parse(fna, "fasta") if seq_record.id == exonerate_target_id][0]
                        SeqIO.write(sseq_seq, sseq_seq_faa, "fasta")
                exonerate_output = os.path.join(output_folder, "exonerate_output_%s_vs_%s.txt"%(query_seq_id, exonerate_target_id))
                exonerate_result = runExonerate(query_seq_faa, sseq_seq_faa, exonerate_output, debug=debug)
                current_result.update(exonerate_result)
        analysis_result.append(current_result)

    return analysis_result

def analysisOutput(analysis_result, analysis_output):
    """
    """
    analysis_header = ["query_seq_id", "blastp_sseqid", "blastp_evalue", "blastp_bitscore",  "tblastn_sseqid", "tblastn_evalue", "tblastn_bitscore", "exonerate_score", "exonerate_hit_range"]
    with open(analysis_output,"w") as csvfile:
        dict_writer = csv.DictWriter(csvfile, fieldnames=analysis_header, delimiter="\t")
        dict_writer.writeheader()            
        dict_writer.writerows(analysis_result)    

if __name__ == "__main__":
    main()    

